import torch
from torch import nn


class SingleNeuronModel(nn.Module):
    def __init__(self, model, idx):
        super().__init__()
        self.model = model
        self.idx = idx

    def forward(self, x):
        return self.model(x)[:, self.idx].squeeze()


def infer_device(module):
    for param in module.parameters():
        return param.device
    for buffer in module.buffers():
        return buffer.device
    return torch.device("cpu")



from contextlib import contextmanager

import numpy as np
import torch

from .training import eval_state


def get_module_output(model, input_shape, use_cuda=True):
    """
    Return the output shape of the model when fed in an array of `input_shape`.
    Note that a zero array of shape `input_shape` is fed into the model and the
    shape of the output of the model is returned.

    Args:
        model (nn.Module): PyTorch module for which to compute the output shape
        input_shape (tuple): Shape specification for the input array into the model
        use_cuda (bool, optional): If True, model will be evaluated on CUDA if available. Othewrise
            model evaluation will take place on CPU. Defaults to True.

    Returns:
        tuple: output shape of the model

    """
    # infer the original device
    initial_device = next(iter(model.parameters())).device
    device = "cuda" if torch.cuda.is_available() and use_cuda else "cpu"
    with eval_state(model):
        with torch.no_grad():
            input = torch.zeros(1, *input_shape[1:], device=device)
            output = model.to(device)(input)
    model.to(initial_device)
    return output.shape


@contextmanager
def no_transforms(dat):
    """
    Contextmanager for the dataset object. It temporarily removes the transforms.
    Args:
        dat: Dataset object. Either FileTreeDataset or StaticImageSet

    Yields: The dataset object without transforms
    """
    transforms = dat.transforms
    try:
        dat.transforms = []
        yield dat
    finally:
        dat.transforms = transforms


def anscombe(x):
    """Compute Anscombe transform."""
    return 2 * np.sqrt(x + 3 / 8)



import logging

import torch
from torch import nn
from torch.nn import ModuleDict
from torch.nn.init import xavier_normal

from .base import Shifter

logger = logging.getLogger(__name__)


class MLP(Shifter):
    def __init__(self, input_features=2, hidden_channels=10, shift_layers=1, **kwargs):
        """
        Multi-layer perceptron shifter
        Args:
            input_features (int): number of input features, defaults to 2.
            hidden_channels (int): number of hidden units.
            shift_layers(int): number of shifter layers (n=1 will correspond to a network without a hidden layer).
            **kwargs:
        """
        super().__init__()

        prev_output = input_features
        feat = []
        for _ in range(shift_layers - 1):
            feat.extend([nn.Linear(prev_output, hidden_channels), nn.Tanh()])
            prev_output = hidden_channels

        feat.extend([nn.Linear(prev_output, 2), nn.Tanh()])
        self.mlp = nn.Sequential(*feat)

    def regularizer(self):
        return 0

    def initialize(self):
        for linear_layer in [p for p in self.parameters() if isinstance(p, nn.Linear)]:
            xavier_normal(linear_layer.weight)

    def forward(self, pupil_center, trial_idx=None):
        if trial_idx is not None:
            pupil_center = torch.cat((pupil_center, trial_idx), dim=1)
        if not self.mlp[0].in_features == pupil_center.shape[1]:
            raise ValueError(
                "The expected input shape of the shifter and the shape of the input do not match! "
                "(Maybe due to the appending of trial_idx to pupil_center?)"
            )
        return self.mlp(pupil_center)


class MLPShifter(ModuleDict):
    def __init__(
        self, data_keys, input_channels=2, hidden_channels_shifter=2, shift_layers=1, gamma_shifter=0, **kwargs
    ):
        """
        Args:
            data_keys (list of str): keys of the shifter dictionary, correspond to the data_keys of the nnfabirk dataloaders
            gamma_shifter: weight of the regularizer

            See docstring of base class for the other arguments.
        """
        super().__init__()
        self.gamma_shifter = gamma_shifter
        for k in data_keys:
            self.add_module(k, MLP(input_channels, hidden_channels_shifter, shift_layers))

    def initialize(self, **kwargs):
        logger.info("Ignoring input {} when initializing {}".format(repr(kwargs), self.__class__.__name__))
        for linear_layer in [p for p in self.parameters() if isinstance(p, nn.Linear)]:
            xavier_normal(linear_layer.weight)

    def regularizer(self, data_key):
        return self[data_key].regularizer() * self.gamma_shifter



import json
from collections import namedtuple
from datetime import datetime
from pathlib import Path
from zipfile import ZipFile

import h5py
import numpy as np
from scipy.signal import convolve2d
from torch.utils.data import Dataset

from ..utils import no_transforms, recursively_load_dict_contents_from_group
from .exceptions import DoesNotExistException, InconsistentDataException
from .transforms import (
    DataTransform,
    Delay,
    Invertible,
    MovieTransform,
    StaticTransform,
    Subsequence,
)
from .utils import convert_static_h5_dataset_to_folder, zip_dir


class AttributeHandler:
    def __init__(self, name, h5_handle):
        """
        Can be used to turn a dataset within a hdf5 dataset into an attribute.
        Args:
            name:       name of the dataset in the hdf5 file
            h5_handle:  file handle for the hdf5 file
        """
        assert name in h5_handle, "{} must be in {}".format(name, h5_handle)
        self.name = name
        self.h5_handle = h5_handle

    def __getattr__(self, item):
        if item in self.h5_handle[self.name]:
            ret = self.h5_handle[self.name][item][()]
            if ret.dtype.char == "S":  # convert bytes to unicode
                ret = ret.astype(str)
            return ret
        else:
            raise AttributeError("Attribute {} not found".format(item))

    def __getitem__(self, item):
        return getattr(self, item)

    def keys(self):
        return self.h5_handle[self.name].keys()

    def __dir__(self):
        attrs = set(super().__dir__())
        return attrs.union(set(self.h5_handle[self.name].keys()))


class AttributeTransformer(AttributeHandler):
    def __init__(self, name, h5_handle, transforms, data_group):
        """
        Allows for id_transform of transforms to be applied to the
        specified attribute. Otherwise behaves like an AttributeHandler
        Args:
            name:       see AttributeHandler
            h5_handle:  see AttributeHandler
            transforms: the set of transforms that's supposed to be applied
            data_group: the data_key of the dataset that this attribute represents
        """
        super().__init__(name, h5_handle)
        self.transforms = transforms
        self.data_group = data_group

    def __getattr__(self, item):
        ret = {self.data_group: super().__getattr__(item)}
        for tr in self.transforms:
            if hasattr(tr, "id_transform"):
                ret = tr.id_transform(ret)

        return ret[self.data_group]


default_image_datapoint = namedtuple("DefaultDataPoint", ["images", "responses"])


class TransformDataset(Dataset):
    def __init__(self, *data_keys, transforms=None):
        """
        Abstract Class for Datasets with transformations, providing `transform` and `invert` functions
        to apply data transformation on the elements.
        Args:
            transforms: list of transforms to be applied to each data point
        """
        self.transforms = transforms or []

        self.data_keys = data_keys
        if set(data_keys) == {"images", "responses"}:
            # this version IS serializable in pickle
            self.data_point = default_image_datapoint
        else:
            # this version is NOT - you cannot use this with a dataloader with num_workers > 1
            self.data_point = namedtuple("DataPoint", data_keys)

    def transform(self, x, exclude=None):
        """
        Apply transform on a data element from the dataset
        Args:
            x (tuple): a data element from the dataset
            exclude (Transform, optional): Type of data transformer to be excluded from transform list. Defaults to None.
        Returns:
            tuple: transformed data element
        """

        for tr in self.transforms:
            if exclude is None or not isinstance(tr, exclude):
                x = tr(x)
        return x

    def invert(self, x, exclude=None):
        for tr in reversed(filter(lambda tr: not isinstance(tr, exclude), self.transforms)):
            if not isinstance(tr, Invertible):
                raise TypeError("Cannot invert", tr.__class__.__name__)
            else:
                x = tr.inv(x)
        return x

    def __iter__(self):
        yield from map(self.__getitem__, range(len(self)))

    def __repr__(self):
        return (
            "{} m={}:\n\t({})".format(self.__class__.__name__, len(self), ", ".join(self.data_keys))
            + "\n\t[Transforms: "
            + "->".join([repr(tr) for tr in self.transforms])
            + "]"
        )


class H5SequenceSet(TransformDataset):
    def __init__(self, filename, *data_keys, output_rename=None, transforms=None, output_dict=False):
        super().__init__(transforms=transforms)

        self.output_dict = output_dict

        if output_rename is None:
            output_rename = {}

        # a flag that can be changed to turn renaming on/off
        self.rename_output = True

        self.output_rename = output_rename

        self._fid = h5py.File(filename, "r")
        self.data = self._fid
        self.data_loaded = False

        # ensure that all elements of data_keys exist
        m = None
        for key in data_keys:
            assert key in self.data, "Could not find {} in file".format(key)
            l = len(self.data[key])
            if m is not None and l != m:
                raise ValueError("groups have different length")
            m = l
        self._len = m

        # Specify which types of transforms are accepted
        self._transform_set = DataTransform

        self.data_keys = data_keys
        self.transforms = transforms or []

        self.data_point = namedtuple("DataPoint", data_keys)
        renamed_keys = [output_rename.get(k, k) for k in data_keys]
        self.output_point = namedtuple("OutputPoint", renamed_keys)

    def __dir__(self):
        attrs = set(super().__dir__())
        return attrs.union(set(self._fid.keys()))

    def load_content(self):
        self.data = recursively_load_dict_contents_from_group(self._fid)
        self.data_loaded = True

    def unload_content(self):
        self.data = self._fid
        self.data_loaded = False

    def __len__(self):
        return self._len

    def __getitem__(self, item):
        x = self.data_point(*(np.array(self.data[g][item if self.data_loaded else str(item)]) for g in self.data_keys))
        for tr in self.transforms:
            assert isinstance(tr, self._transform_set)
            x = tr(x)

        # convert to output point
        if self.rename_output:
            x = self.output_point(*x)

        if self.output_dict:
            x = x._asdict()
        return x

    def __getattr__(self, item):
        if item in self.data:
            item = self.data[item]
            if isinstance(item, h5py.Dataset):
                dtype = item.dtype
                item = item[()]
                if dtype.char == "S":  # convert bytes to unicode
                    item = item.astype(str)
                return item
            return item
        else:
            # TODO: check for a proper way to handle cases where super doesn't have __getattr__
            return super().__getattr__(item)

    def __repr__(self):
        names = ["{} -> {}".format(k, self.output_rename[k]) if k in self.output_rename else k for k in self.data_keys]
        s = "{} m={}:\n\t({})".format(self.__class__.__name__, len(self), ", ".join(names))
        if self.transforms is not None:
            s += "\n\t[Transforms: " + "->".join([repr(tr) for tr in self.transforms]) + "]"
        return s


class MovieSet(H5SequenceSet):
    """
    Extension to H5SequenceSet with specific HDF5 dataset assumed. Specifically,
    it assumes that properties such as `neurons` and `stats` are present in the dataset.
    """

    def __init__(self, filename, *data_groups, output_rename=None, transforms=None, stats_source="all"):
        super().__init__(filename, *data_groups, output_rename=output_rename, transforms=transforms)
        self.stats_source = stats_source

        # set to accept only MovieTransform
        self._transform_set = MovieTransform

    @property
    def neurons(self):
        return AttributeTransformer("neurons", self.data, self.transforms, data_group="responses")

    @property
    def n_neurons(self):
        return len(self.neurons.unit_ids)

    @property
    def input_shape(self):
        name = self.output_rename.get("inputs", "inputs") if self.rename_output else "inputs"
        return (1,) + getattr(self[0], name).shape

    def transformed_mean(self, stats_source=None):
        if stats_source is None:
            stats_source = self.stats_source

        tmp = [np.atleast_1d(self.statistics[g][stats_source]["mean"][()]) for g in self.data_keys]
        x = self.transform(self.data_point(*tmp), exclude=(Subsequence, Delay))
        if self.rename_output:
            x = self.output_point(*x)
        return x

    def rf_base(self, stats_source="all"):
        N, c, t, w, h = self.img_shape
        t = min(t, 150)
        mean = lambda dk: self.statistics[dk][stats_source]["mean"][()]
        d = dict(
            inputs=np.ones((1, c, t, w, h)) * np.array(mean("inputs")),
            eye_position=np.ones((1, t, 1)) * mean("eye_position")[None, None, :],
            behavior=np.ones((1, t, 1)) * mean("behavior")[None, None, :],
            responses=np.ones((1, t, 1)) * mean("responses")[None, None, :],
        )
        return self.transform(self.data_point(*[d[dk] for dk in self.data_keys]), exclude=Subsequence)

    def rf_noise_stim(self, m, t, stats_source="all"):
        """
        Generates a Gaussian white noise stimulus filtered with a 3x3 Gaussian filter
        for the computation of receptive fields. The mean and variance of the Gaussian
        noise are set to the mean and variance of the stimulus ensemble.
        The behvavior, eye movement statistics, and responses are set to their respective means.
        Args:
            m: number of noise samples
            t: length in time
        Returns: tuple of input, behavior, eye, and response
        """
        N, c, _, w, h = self.img_shape
        stat = lambda dk, what: self.statistics[dk][stats_source][what][()]
        mu, s = stat("inputs", "mean"), stat("inputs", "std")
        h_filt = np.float64([[1 / 16, 1 / 8, 1 / 16], [1 / 8, 1 / 4, 1 / 8], [1 / 16, 1 / 8, 1 / 16]])
        noise_input = (
            np.stack([convolve2d(np.random.randn(w, h), h_filt, mode="same") for _ in range(m * t * c)]).reshape(
                (m, c, t, w, h)
            )
            * s
            + mu
        )

        mean_beh = np.ones((m, t, 1)) * stat("behavior", "mean")[None, None, :]
        mean_eye = np.ones((m, t, 1)) * stat("eye_position", "mean")[None, None, :]
        mean_resp = np.ones((m, t, 1)) * stat("responses", "mean")[None, None, :]

        d = dict(
            inputs=noise_input.astype(np.float32),
            eye_position=mean_eye.astype(np.float32),
            behavior=mean_beh.astype(np.float32),
            responses=mean_resp.astype(np.float32),
        )

        return self.transform(
            self.data_point(*[d[dk] for dk in self.data_groups.values()]), exclude=(Subsequence, Delay)
        )


class StaticSet(TransformDataset):
    def __init__(self, *data_keys, transforms=None):
        """
        Abstract class for static datasets. Defines data_keys and a corresponding datapoint.
        """
        super().__init__(transforms=transforms)

        self.data_keys = data_keys
        if set(data_keys) == {"images", "responses"}:
            # this version IS serializable in pickle
            self.data_point = default_image_datapoint
        else:
            # this version is NOT - you cannot use this with a dataloader with num_workers > 1
            self.data_point = namedtuple("DataPoint", data_keys)


class H5ArraySet(StaticSet):
    def __init__(self, filename, *data_keys, transforms=None):
        """
        Dataset for static data stored in hdf5 files.
        Args:
            filename:      filename of the hdf5 file
            *data_keys:    data keys to be read from the file
            transforms:    list of transforms applied to each datapoint
        """
        super().__init__(*data_keys, transforms=transforms)

        self._fid = h5py.File(filename, "r")
        self.data = self._fid
        self.data_loaded = False
        m = None
        for key in data_keys:
            assert key in self.data, "Could not find {} in file".format(key)
            if m is None:
                m = len(self.data[key])
            else:
                assert m == len(self.data[key]), "Length of datasets do not match"
        self._len = m

    def load_content(self):
        self.data = recursively_load_dict_contents_from_group(self._fid)
        self.data_loaded = True

    def unload_content(self):
        self.data = self._fid
        self.data_loaded = False

    def __getitem__(self, item):
        x = self.data_point(*(self.data[g][item] for g in self.data_keys))
        for tr in self.transforms:
            assert isinstance(tr, StaticTransform)
            x = tr(x)
        return x

    def __iter__(self):
        yield from map(self.__getitem__, range(len(self)))

    def __len__(self):
        return self._len

    def __repr__(self):
        return "\n".join(
            ["Tensor {}: {} ".format(key, self.data[key].shape) for key in self.data_keys]
            + ["Transforms: " + repr(self.transforms)]
        )

    def __getattr__(self, item):
        if item in self.data:
            item = self.data[item]
            if isinstance(item, h5py.Dataset):
                dtype = item.dtype
                item = item[()]
                if dtype.char == "S":  # convert bytes to univcode
                    item = item.astype(str)
                return item
            return item
        else:
            raise AttributeError("Item {} not found in {}".format(item, self.__class__.__name__))


class StaticImageSet(H5ArraySet):
    def __init__(self, filename, *data_keys, transforms=None, cache_raw=False, stats_source=None):
        """
        Dataset for h5 files.
        Args:
            filename:       filename of the hdf5 file
            *data_keys:     datasets to be extracted
            transforms:     transforms applied to each data point
            cache_raw:      whether to cache the raw (untransformed) datapoints
            stats_source:   statistic source to be used.
        """
        super().__init__(filename, *data_keys, transforms=transforms)
        self.cache_raw = cache_raw
        self.last_raw = None
        self.stats_source = stats_source if stats_source is not None else "all"

    @property
    def n_neurons(self):
        return len(self[0].responses)

    @property
    def neurons(self):
        return AttributeTransformer("neurons", self.data, self.transforms, data_group="responses")

    @property
    def info(self):
        return AttributeHandler("item_info", self.data)

    @property
    def img_shape(self):
        return (1,) + self[0].images.shape

    def transformed_mean(self, stats_source=None):
        if stats_source is None:
            stats_source = self.stats_source

        tmp = [np.atleast_1d(self.statistics[dk][stats_source]["mean"][()]) for dk in self.data_keys]
        return self.transform(self.data_point(*tmp))

    def __repr__(self):
        return super().__repr__() + (
            "\n\t[Stats source: {}]".format(self.stats_source) if self.stats_source is not None else ""
        )

    def __dir__(self):
        attrs = set(self.__dict__).union(set(dir(type(self))))
        return attrs.union(set(self.data.keys()))


class DirectoryAttributeHandler:
    def __init__(self, path, links=None):
        """
        Class that can be used to represent a subdirectory of a FileTree as a property in a FileTree dataset.
        Args:
            path (pathlib.Path object): path to the subdiretory
            links (dict, optional): rename mapping for entries within the `path`. Defaults to None, in which case
                no name mapping is performed when fetching the attribute.
        """
        self.links = links or {}
        self.path = path

    def __getattr__(self, item):
        item_path = self.resolve_item_path(item)

        if item_path.exists() and item_path.is_dir():
            val = DirectoryAttributeHandler(item_path, links=self.links)
        else:
            data_path = item_path.with_suffix(".npy")
            if data_path.exists() and data_path.is_file():
                val = np.load(data_path)
            else:
                raise AttributeError("Attribute {} not found".format(item))
        return val

    def resolve_item_path(self, item):
        """
        Formulates a path to the `item`, taken relative to the
        `path` attribute of this object. If the `item` has an entry in
        `links` dictionary, then the name is mapped into that instead.

        Args:
            item (str): Name of item to obtain.

        Returns:
            pathlib.Path object: Formulated full path to the target item
        """
        if item in self.links:
            item = self.links[item]
        return self.path / item

    def __getitem__(self, item):
        return getattr(self, item)

    def keys(self):
        return [e.stem for e in self.path.glob("*")]

    def __dir__(self):
        attrs = set(super().__dir__())
        return attrs.union(set(self.keys())).union(set(self.links.keys()))


class DirectoryAttributeTransformer(DirectoryAttributeHandler):
    def __init__(self, path, transforms, data_group, links=None):
        """
        Class that can be used to represent a subdirectory of a FileTree as a property in a FileTree dataset.
        Like DirectoryAttributeHandler but allows for id_transform of transforms to be applied to the
        specified attribute.
        Args:
            path (pathlib.Path object): path to the subdiretory
            transforms (list): A list of DataTransform objects, whose `id_transform` will be applied to the loaded property sequentially
            data_group (str): Name of data_group that the transforms should be applied as
            links (dict, optional): rename mapping for entries within the `path`. Defaults to None, in which case
                no name mapping is performed when fetching the attribute.
        """

        super().__init__(path, links=links)
        self.transforms = transforms
        self.data_group = data_group

    def __getattr__(self, item):
        ret = {self.data_group: super().__getattr__(item)}
        for tr in self.transforms:
            ret = tr.id_transform(ret)
        return ret[self.data_group]


class FileTreeDatasetBase(TransformDataset):
    _default_config = {"links": {}}
    # specify list of transform types that are acceptable
    _transform_types = (DataTransform,)

    def __init__(self, dirname, *data_keys, transforms=None, use_cache=True, output_rename=None, output_dict=False):
        """
        Dataset stored as a file tree. The tree needs to have the subdirs data, meta, meta/neurons, meta/statistics,
        and meta/trials. Please refer to convert_static_h5_dataset_to_folder in neuralpredictors.data.utils for an
        example on how to export an hdf5 file into folder structure compatible with this dataset.


        Here is an example. Data directories with too many entries have trials as .npy files
        named 0.npy, 1.npy, ...
        The meta/trials subdirectory must have single .npy files with arrays that provide additional trial based
        meta data.
        static22564-2-13-preproc0
        ├── data
        │   ├── behavior [directory with 5955 entries]
        │   ├── images [directory with 5955 entries]
        │   ├── pupil_center [5955 entries]
        │   └── responses [5955 entries]
        └── meta
            ├── neurons
            │   ├── animal_ids.npy
            │   ├── area.npy
            │   ├── layer.npy
            │   ├── scan_idx.npy
            │   ├── sessions.npy
            │   └── unit_ids.npy
            ├── statistics
            │   ├── behavior
            │   │   ├── all
            │   │   │   ├── max.npy
            │   │   │   ├── mean.npy
            │   │   │   ├── median.npy
            │   │   │   ├── min.npy
            │   │   │   └── std.npy
            │   │   └── stimulus_frame
            │   │       ├── max.npy
            │   │       ├── mean.npy
            │   │       ├── median.npy
            │   │       ├── min.npy
            │   │       └── std.npy
            │   ├── images
            │   │   ├── all
            │   │   │   ├── max.npy
            │   │   │   ├── mean.npy
            │   │   │   ├── median.npy
            │   │   │   ├── min.npy
            │   │   │   └── std.npy
            │   │   └── stimulus_frame
            │   │       ├── max.npy
            │   │       ├── mean.npy
            │   │       ├── median.npy
            │   │       ├── min.npy
            │   │       └── std.npy
            │   ├── pupil_center
            │   │   ├── all
            │   │   │   ├── max.npy
            │   │   │   ├── mean.npy
            │   │   │   ├── median.npy
            │   │   │   ├── min.npy
            │   │   │   └── std.npy
            │   │   └── stimulus_frame
            │   │       ├── max.npy
            │   │       ├── mean.npy
            │   │       ├── median.npy
            │   │       ├── min.npy
            │   │       └── std.npy
            │   └── responses
            │       ├── all
            │       │   ├── max.npy
            │       │   ├── mean.npy
            │       │   ├── median.npy
            │       │   ├── min.npy
            │       │   └── std.npy
            │       └── stimulus_frame
            │           ├── max.npy
            │           ├── mean.npy
            │           ├── median.npy
            │           ├── min.npy
            │           └── std.npy
            └── trials [12 entries]

        Args:
            dirname:     root directory name
            *data_keys:  data items to be extraced (must be subdirectories of root/data)
            transforms:  transforms to be applied to the data (see TransformDataset)
        """
        super().__init__(*data_keys, transforms=transforms)

        number_of_files = []

        self.output_dict = output_dict
        self.use_cache = use_cache

        if output_rename is None:
            output_rename = {}

        # a flag that can be changed to turn renaming on/off
        self.rename_output = bool(output_rename)
        self._output_rename = output_rename
        renamed_keys = [output_rename.get(k, k) for k in data_keys]
        self._output_point = namedtuple("OutputPoint", renamed_keys) if output_rename else self.data_point

        # if dirname is a zip file, auto expand into the container folder
        if dirname.endswith(".zip"):
            if not Path(dirname[:-4]).exists():
                self.unzip(dirname, Path(dirname).absolute().parent)
            else:
                print("{} exists already. Not unpacking {}".format(dirname[:-4], dirname))
            dirname = dirname[:-4]

        self.dirname = dirname
        self.basepath = Path(dirname).absolute()
        self._config_file = self.basepath / "config.json"

        # if no config file, create one based on default config
        if not self._config_file.exists():
            self._save_config(self._default_config)

        # verify that valid data path exists for each data_key and count number of files
        for data_key in data_keys:
            if data_key not in self.trial_info.keys():
                datapath = self.resolve_data_path(data_key)
                number_of_files.append(len(list(datapath.glob("*"))))
            else:
                number_of_files.append(len(self.trial_info[data_key]))

        # verify that all data_keys directories contain identical number of files (= number of data points)
        if not np.all(np.diff(number_of_files) == 0):
            raise InconsistentDataException("Number of data points is not equal")
        # keep the number of files as the number of data points
        self._len = number_of_files[0]

        self._cache = {data_key: {} for data_key in data_keys}

    def resolve_data_path(self, data_key):
        """
        Given a data_key, resolves the folder within self.basepath/data directory. If relevant "links"
        entry exists in the config, the name mapping is performed. Finally, the resultant path is checked
        for validness (is it a directory that exists) and raises an exception if not found. Otherwise, the
        resultant path object is returned.

        Args:
            data_key (str): data_key to find corresponding subdirectory under `self.basepath/data`

        Raises:
            DoesNotExistException: If the target path is not a valid directory, this exception is raised.

        Returns:
            pathlib.Path object: Valid directory path to the target data_group
        """
        if self.link_exists(data_key):
            data_key = self.config["links"][data_key]
        datapath = self.basepath / "data" / data_key

        if not datapath.exists() or not datapath.is_dir():
            raise DoesNotExistException("Data path {} is not a valid directory".format(datapath))
        return datapath

    @staticmethod
    def unzip(filename, path):
        """
        Unzips the target file with `filename` into the specified `path`

        Args:
            filename (str): Path to the zip file
            path (str): Path to expand the zip content into
        """
        print("Unzipping {} into {}".format(filename, path))
        with ZipFile(filename, "r") as zip_obj:
            zip_obj.extractall(path)

    def link_exists(self, link):
        """
        Checks if an entry for `link` exists in the "links" config

        Args:
            link (str): data_group name to check for an entry in "links"

        Returns:
            bool: True if a relevant entry is found in "links" config
        """
        return "links" in self.config and link in self.config["links"]

    @property
    def config(self):
        with open(self._config_file) as fid:
            return json.load(fid)

    def _save_config(self, cfg):
        with open(self._config_file, "w") as fid:
            return json.dump(cfg, fid)

    def __len__(self):
        return self._len

    def __getitem__(self, item):
        # load data from cache or disk
        ret = []
        for data_key in self.data_keys:
            if self.use_cache and item in self._cache[data_key]:
                ret.append(self._cache[data_key][item])
            else:
                if data_key in self.trial_info.keys():
                    val = self.trial_info[data_key][item : item + 1]
                else:
                    datapath = self.resolve_data_path(data_key)
                    val = np.load(datapath / "{}.npy".format(item))
                if self.use_cache:
                    self._cache[data_key][item] = val
                ret.append(val)

        # create data point and transform
        x = self.data_point(*ret)

        for tr in self.transforms:
            # ensure only specified types of transforms are used
            assert isinstance(tr, self._transform_types)
            x = tr(x)

        # apply output rename if necessary
        if self.rename_output:
            x = self._output_point(*x)

        if self.output_dict:
            x = x._asdict()

        return x

    def add_log_entry(self, msg):
        """
        Add a new log entry `msg` into the "change.log" file. The message will be timestamped

        Args:
            msg (str): Message to be logged
        """
        timestamp = datetime.now().strftime("%d-%b-%Y (%H:%M:%S.%f)")
        with open(self.basepath / "change.log", "a+") as fid:
            fid.write("{}: {}\n".format(timestamp, msg))

    @property
    def change_log(self):
        """
        Convenience property to print the content of change.log file if it exists
        """
        if (self.basepath / "change.log").exists():
            with open(self.basepath / "change.log", "r") as fid:
                print("".join(fid.readlines()))

    def zip(self, filename=None):
        """
        Zips current dataset.

        Args:
            filename:  Filename for the zip. Directory name + zip by default.
        """

        if filename is None:
            filename = str(self.basepath) + ".zip"
        zip_dir(filename, self.basepath)

    def add_link(self, attr, new_name):
        """
        Add a new dataset that links to an existing dataset.

        For instance `targets` that links to `responses`

        Args:
            attr:       existing attribute such as `responses`
            new_name:   name of the new attribute reference.
        """
        if not (self.basepath / "data/{}".format(attr)).exists():
            raise DoesNotExistException("Link target does not exist")

        if (self.basepath / "data/{}".format(new_name)).exists():
            raise FileExistsError("Link target already exists")

        config = self.config
        if not "links" in config:
            config["links"] = {}
        config["links"][new_name] = attr
        self._save_config(config)

    @property
    def neurons(self):
        return DirectoryAttributeTransformer(
            self.basepath / "meta/neurons",
            self.transforms,
            data_group="responses" if "responses" in self.data_keys else "targets",
        )

    @property
    def trial_info(self):
        return DirectoryAttributeHandler(self.basepath / "meta/trials")

    @property
    def statistics(self):
        return DirectoryAttributeHandler(self.basepath / "meta/statistics", self.config["links"])

    @staticmethod
    def match_order(target, permuted, not_exist_ok=False):
        """
        Matches the order or rows in permuted to by returning an index array such that.
        Args:
            not_exist_ok: if the element does not exist, don't return an index
        Returns: index array `idx` such that `target == permuted[idx, :]`
        """

        order, target_idx = [], []
        unmatched_counter = 0
        for i, row in enumerate(target):
            idx = np.sum(permuted - row, axis=1) == 0
            if not not_exist_ok:
                assert idx.sum() == 1
            if idx.sum() == 1:
                order.append(np.where(idx)[0][0])
                target_idx.append(i)
            else:
                unmatched_counter += 1
        if not_exist_ok:
            print("Encountered {} unmatched elements".format(unmatched_counter))
        return np.array(target_idx, dtype=int), np.array(order, dtype=int)

    def add_neuron_meta(self, name, animal_id, session, scan_idx, unit_id, values, fill_missing=None):
        """
        Add new meta information about neurons.
        Args:
            name:       name of the new meta information
            animal_id:  array with animal_ids matching the size of the first dimension of values
            session:    array with session matching the size of the first dimension of values
            scan_idx:   array with scan_idx matching the size of the first dimension of values
            unit_id:    array with unit_id matching the size of the first dimension of values
            values:     new meta information. First dimension must correspond to neurons.
            fill_missing: fill the values of the new attribute with the specified value. Defaults to None,
                in which case missing values are filled with NaN
        """

        with no_transforms(self):

            if not self.n_neurons == len(values):
                raise InconsistentDataException(
                    f"Number of values ({len(values)}) and neurons in the datasets ({self.n_neurons}) is not consistent."
                )

        if not len(animal_id) == len(session) == len(scan_idx) == len(unit_id) == len(values):
            raise InconsistentDataException("number of trials and identifiers not consistent")

        target = np.c_[(self.neurons.animal_ids, self.neurons.sessions, self.neurons.scan_idx, self.neurons.unit_ids)]
        permuted = np.c_[(animal_id, session, scan_idx, unit_id)]

        vals = np.ones((len(target),) + values.shape[1:], dtype=values.dtype) * (
            np.nan if fill_missing is None else fill_missing
        )
        tidx, idx = self.match_order(target, permuted, not_exist_ok=fill_missing is not None)

        assert np.sum(target[tidx] - permuted[idx, ...]) == 0, "Something went wrong in sorting"

        vals[tidx, ...] = values[idx, ...]
        np.save(self.basepath / "meta/neurons/{}.npy".format(name), vals)
        self.add_log_entry("Added new neuron meta attribute {} to meta/neurons".format(name))

    def __repr__(self):
        return "{} {} (n={} items)\n\t{}".format(
            self.__class__.__name__, self.basepath, self._len, ", ".join(self.data_keys)
        )


class FileTreeDataset(FileTreeDatasetBase):
    _transform_types = (StaticTransform,)

    @staticmethod
    def initialize_from(filename, outpath=None, overwrite=False):
        """
        Convenience function. See `convert_static_h5_dataset_to_folder` in `.utils`
        """
        convert_static_h5_dataset_to_folder(filename, outpath=outpath, overwrite=overwrite)

    @property
    def img_shape(self):
        return (1,) + self[0].images.shape

    @property
    def n_neurons(self):
        target_group = "responses" if "responses" in self.data_keys else "targets"
        val = self[0]
        if hasattr(val, target_group):
            val = getattr(val, target_group)
        else:
            val = val[target_group]
        return len(val)

    def change_log(self):
        if (self.basepath / "change.log").exists():
            with open(self.basepath / "change.log", "r") as fid:
                print("".join(fid.readlines()))

    def zip(self, filename=None):
        """
        Zips current dataset.
        Args:
            filename:  Filename for the zip. Directory name + zip by default.
        """

        if filename is None:
            filename = str(self.basepath) + ".zip"
        zip_dir(filename, self.basepath)

    def unzip(self, filename, path):
        print("Unzipping {} into {}".format(filename, path))
        with ZipFile(filename, "r") as zip_obj:
            zip_obj.extractall(path)

    def add_link(self, attr, new_name):
        """
        Add a new dataset that links to an existing dataset.
        For instance `targets` that links to `responses`
        Args:
            attr:       existing attribute such as `responses`
            new_name:   name of the new attribute reference.
        """
        if not (self.basepath / "data/{}".format(attr)).exists():
            raise DoesNotExistException("Link target does not exist")


class MovieFileTreeDataset(FileTreeDatasetBase):
    _transform_types = (MovieTransform,)

    def __init__(self, *args, stats_source=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.stats_source = stats_source if stats_source is not None else "all"

    # the followings are provided for compatibility with MovieSet
    @property
    def types(self):
        return self.trial_info.types

    @property
    def tiers(self):
        return self.trial_info.tiers

    @property
    def n_neurons(self):
        target_group = "responses" if "responses" in self.data_keys else "targets"
        # check if output has been renamed
        if self.rename_output:
            target_group = self._output_rename.get(target_group, target_group)

        val = self[0]
        if hasattr(val, target_group):
            val = getattr(val, target_group)
        else:
            val = val[target_group]
        return val.shape[-1]

    @property
    def statistics(self):
        return DirectoryAttributeHandler(self.basepath / "meta/statistics", self.config["links"])

    def transformed_mean(self, stats_source=None):
        if stats_source is None:
            stats_source = self.stats_source

        tmp = [np.atleast_1d(self.statistics[g][stats_source]["mean"][()]) for g in self.data_keys]
        x = self.transform(self.data_point(*tmp), exclude=(Subsequence, Delay))
        if self.rename_output:
            x = self._output_point(*x)
        return x



import torch


def cosine_similarity(tensor, mean_value=None):
    # if mean_value is not None:
    #     tensor = tensor - mean_value
    # else:
    #     tensor = tensor - tensor.mean(dim=-1, keepdim=True)
    tensor_normed = tensor / torch.norm(tensor, p=2, dim=1, keepdim=True)
    return tensor_normed @ tensor_normed.T


def eucl_dist_similarity(x):
    # normalize
    x = x - x.mean(dim=-1, keepdim=True)
    x = x / x.norm(dim=-1, keepdim=True)

    # make distance
    x = torch.cdist(x, x)
    x = -(x - 1)
    return x


def dot_product_similarity(x):
    x_normed = x / x.norm(dim=1, p=2, keepdim=True)
    return x_normed @ x_normed.T



from torch import nn


class SqueezeExcitationBlock(nn.Module):
    def __init__(self, in_ch, reduction=16):
        """
        A squeeze and excitation block as proposed by https://arxiv.org/abs/1709.01507
        Args:
            in_ch (int): number of input channels
            reduction (int): reduction factor to calculate the output channels.
        """
        super().__init__()
        self.se = nn.Sequential(
            GlobalAvgPool(),
            nn.Linear(in_ch, in_ch // reduction),
            nn.ReLU(inplace=True),
            nn.Linear(in_ch // reduction, in_ch),
            nn.Sigmoid(),
        )

    def forward(self, x):
        se_weight = self.se(x).unsqueeze(-1).unsqueeze(-1)
        return x.mul(se_weight)


class GlobalAvgPool(nn.Module):
    def __init__(self):
        """
        Helper class used by the SqueezeExcitationBlock
        """
        super().__init__()

    def forward(self, x):
        return x.view(*(x.shape[:-2]), -1).mean(-1)



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init


class AttentionConv(nn.Module):
    """
    Implementation adapted from https://github.com/leaderj1001/Stand-Alone-Self-Attention
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        groups=1,
        bias=False,
    ):
        """
        Parameters are intended to behave equivalently (and therefore sever as a drop-in replacement) to `torch.Conv2d`.
        Nevertheless, the underlying mechanism is conceptually different.
        Refer to https://arxiv.org/pdf/1906.05909.pdf for more information.
        """
        super().__init__()
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.groups = groups

        assert (
            self.out_channels % self.groups == 0
        ), "out_channels should be divided by groups. (example: out_channels: 40, groups: 4)"

        self.rel_h = nn.Parameter(torch.randn(out_channels // 2, 1, 1, kernel_size, 1), requires_grad=True)
        self.rel_w = nn.Parameter(torch.randn(out_channels // 2, 1, 1, 1, kernel_size), requires_grad=True)

        self.key_conv = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=bias)
        self.query_conv = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=bias)
        self.value_conv = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=bias)

        self.reset_parameters()

    def forward(self, x):
        batch, channels, height, width = x.size()

        padded_x = F.pad(x, [self.padding, self.padding, self.padding, self.padding])
        q_out = self.query_conv(x)
        k_out = self.key_conv(padded_x)
        v_out = self.value_conv(padded_x)

        k_out = k_out.unfold(2, self.kernel_size, self.stride).unfold(3, self.kernel_size, self.stride)
        v_out = v_out.unfold(2, self.kernel_size, self.stride).unfold(3, self.kernel_size, self.stride)

        k_out_h, k_out_w = k_out.split(self.out_channels // 2, dim=1)
        k_out = torch.cat((k_out_h + self.rel_h, k_out_w + self.rel_w), dim=1)

        k_out = k_out.contiguous().view(batch, self.groups, self.out_channels // self.groups, height, width, -1)
        v_out = v_out.contiguous().view(batch, self.groups, self.out_channels // self.groups, height, width, -1)

        q_out = q_out.view(batch, self.groups, self.out_channels // self.groups, height, width, 1)

        out = q_out * k_out
        out = F.softmax(out, dim=-1)
        out = torch.einsum("bnchwk,bnchwk -> bnchw", out, v_out).view(batch, -1, height, width)

        return out

    def reset_parameters(self):
        init.kaiming_normal_(self.key_conv.weight, mode="fan_out", nonlinearity="relu")
        init.kaiming_normal_(self.value_conv.weight, mode="fan_out", nonlinearity="relu")
        init.kaiming_normal_(self.query_conv.weight, mode="fan_out", nonlinearity="relu")

        init.normal_(self.rel_h, 0, 1)
        init.normal_(self.rel_w, 0, 1)

    @property
    def weight(self):
        return torch.cat(
            (self.key_conv.weight, self.value_conv.weight, self.query_conv.weight),
            dim=0,
        )



import numpy as np
from .utils.simulated_model import (
    ArbitraryMultiNeuronModel,
    sigmoid,
    generate_gabor_filter,
    ArbitraryNeuronModel,
    ComplexCell,
    generate_gabor_filter_span_as_grid,
)


def neuron1_generator(loc, img_res, transformation_matrix=None, unit_norm=True):
    sigma = 0.13
    sf = 2.5
    gabor_locs = (
        np.array(
            [
                [0.05, 0.05],
                [-0.05, -0.05],
            ]
        )
        + loc
    )
    n_points = 15
    invariance_axis = np.concatenate(
        [np.linspace(-7.5, 7.5, n_points), np.linspace(-7.5, 7.5, n_points)]
    )

    filters = []
    for filter_idx, x in enumerate(invariance_axis):
        gb1 = generate_gabor_filter(
            theta=-np.pi / 2,
            width_x=sigma,
            width_y=sigma,
            frequency=sf,
            phase=np.pi / 2,
            center=gabor_locs[0],
            grid_size=img_res,
            transformation_matrix=transformation_matrix,
        )

        temp = 2
        phase = (sigmoid(x, temp=temp) * 2 - 1) * np.pi / 2
        if filter_idx >= n_points:
            phase = (sigmoid(x, temp=temp) * 2 + 1) * np.pi / 2
        gb2 = generate_gabor_filter(
            theta=np.pi / 8,
            width_x=sigma,
            width_y=sigma,
            frequency=sf,
            phase=phase,
            center=gabor_locs[1],
            grid_size=img_res,
            transformation_matrix=transformation_matrix,
        )

        gb = (gb1 + gb2) / 2

        if unit_norm:
            gb = gb / np.linalg.norm(gb)

        filters.append(gb.astype(np.float32))

    xx, yy = generate_gabor_filter_span_as_grid(
        center=np.array(loc),
        grid_size=img_res,
        transformation_matrix=transformation_matrix,
    )
    grid = xx, yy

    return ArbitraryNeuronModel(np.stack(filters)), grid


def neuron2_generator(loc, img_res, transformation_matrix=None, unit_norm=True, num_filters=30):
    sigma_x = 0.11
    sigma_y = 0.15
    sf = 3
    min_val = 0.9
    max_val = 1.5
    drange = max_val - min_val
    periods = 1

    gabor_locs = np.array([[-0.09, 0.01], [-0.01, -0.05], [0.11, 0.04]]) + loc

    filters = []
    for x in np.linspace(0, 2 * np.pi * periods, periods * num_filters):
        scale1 = (np.sin(x) + 1) / 2 * drange + min_val
        scale2 = (np.sin(x + np.pi / 3) + 1) / 2 * drange + min_val
        scale3 = (np.sin(x + np.pi / 3 * 2) + 1) / 2 * drange + min_val

        gb1 = generate_gabor_filter(
            theta=-np.pi / 8,
            width_x=sigma_x * scale1,
            width_y=sigma_y * scale1,
            frequency=sf / scale1,
            phase=0,
            center=gabor_locs[0],
            grid_size=img_res,
            transformation_matrix=transformation_matrix,
        )

        gb2 = generate_gabor_filter(
            theta=0,
            width_x=sigma_x * scale2,
            width_y=sigma_y * scale2,
            frequency=sf / scale2,
            phase=0,
            center=gabor_locs[1],
            grid_size=img_res,
            transformation_matrix=transformation_matrix,
        )

        gb3 = generate_gabor_filter(
            theta=np.pi / 8,
            width_x=sigma_x * scale3,
            width_y=sigma_y * scale3,
            frequency=sf / scale3,
            phase=0,
            center=gabor_locs[2],
            grid_size=img_res,
            transformation_matrix=transformation_matrix,
        )

        gb = (gb1 + gb2 + gb3) / 3
        # gb = (gb1 + gb3)/3

        if unit_norm:
            gb = gb / np.linalg.norm(gb)

        filters.append(gb.astype(np.float32))

    xx, yy = generate_gabor_filter_span_as_grid(
        center=np.array(loc),
        grid_size=img_res,
        transformation_matrix=transformation_matrix,
    )
    grid = xx, yy

    return ArbitraryNeuronModel(np.stack(filters)), grid


def neuron3_generator(loc, img_res, transformation_matrix=None, unit_norm=True, num_filters=15):
    sigma_x = 0.13
    sigma_y = 0.13
    sf = 2.5
    gabor_locs = (
        np.array(
            [
                [0.2, 0.0],  # right
                [0.0, 0.2],  # top
                [-0.2, 0.0],  # left
                [0.0, -0.2],  # bottom
            ]
        )
        * 0.7
        + loc
    )
    n_points = num_filters
    invariance_axis = np.concatenate(
        [np.linspace(0, 2 * np.pi, n_points)] * len(gabor_locs)
    )

    filters = []
    for filter_idx, x in enumerate(invariance_axis):
        if (filter_idx >= n_points * 0) & (filter_idx <= n_points * 1):
            theta1 = -np.pi / 2 + x
        else:
            theta1 = -np.pi / 2

        gb1 = generate_gabor_filter(
            theta=theta1,
            width_x=sigma_x,
            width_y=sigma_y,
            frequency=sf,
            phase=np.pi / 2,
            center=gabor_locs[0],
            grid_size=img_res,
            transformation_matrix=transformation_matrix,
        )

        if (filter_idx >= n_points * 1) & (filter_idx <= n_points * 2):
            theta2 = x
        else:
            theta2 = 0
        gb2 = generate_gabor_filter(
            theta=theta2,
            width_x=sigma_x,
            width_y=sigma_y,
            frequency=sf,
            phase=np.pi / 2,
            center=gabor_locs[1],
            grid_size=img_res,
            transformation_matrix=transformation_matrix,
        )

        if (filter_idx >= n_points * 2) & (filter_idx <= n_points * 3):
            theta3 = np.pi / 2 + x
        else:
            theta3 = np.pi / 2
        gb3 = generate_gabor_filter(
            theta=theta3,
            width_x=sigma_x,
            width_y=sigma_y,
            frequency=sf,
            phase=np.pi / 2,
            center=gabor_locs[2],
            grid_size=img_res,
            transformation_matrix=transformation_matrix,
        )

        if (filter_idx >= n_points * 3) & (filter_idx <= n_points * 4):
            theta4 = x
        else:
            theta4 = 0
        gb4 = generate_gabor_filter(
            theta=theta4,
            width_x=sigma_x,
            width_y=sigma_y,
            frequency=sf,
            phase=-np.pi / 2,
            center=gabor_locs[3],
            grid_size=img_res,
            transformation_matrix=transformation_matrix,
        )

        gb = (gb1 + gb2 + gb3 + gb4) / 4

        if unit_norm:
            gb = gb / np.linalg.norm(gb)

        filters.append(gb.astype(np.float32))

    xx, yy = generate_gabor_filter_span_as_grid(
        center=np.array(loc),
        grid_size=img_res,
        transformation_matrix=transformation_matrix,
    )
    grid = xx, yy

    return ArbitraryNeuronModel(np.stack(filters)), grid


def phase_invariant_neuron_generator(
    loc, img_res, transformation_matrix=None, unit_norm=True, num_filters=32
):
    sigma_x = 0.14
    sigma_y = 0.14
    sf = 2.3
    phases = np.linspace(0, 2 * np.pi, num_filters, endpoint=False)
    filters = []
    for _, phase in enumerate(phases):
        gb = generate_gabor_filter(
            theta=0,
            width_x=sigma_x,
            width_y=sigma_y,
            frequency=sf,
            phase=phase,
            center=loc,
            grid_size=img_res,
            transformation_matrix=transformation_matrix,
        )
        if unit_norm:
            gb = gb / np.linalg.norm(gb)

        filters.append(gb.astype(np.float32))

    xx, yy = generate_gabor_filter_span_as_grid(
        center=np.array(loc),
        grid_size=img_res,
        transformation_matrix=transformation_matrix,
    )
    grid = xx, yy

    return ArbitraryNeuronModel(np.stack(filters)), grid


def even_gabor_neuron_generator(
    loc, img_res, transformation_matrix=None, unit_norm=True
):
    sigma_x = 0.14
    sigma_y = 0.14
    sf = 2.3
    gb = generate_gabor_filter(
        theta=0,
        width_x=sigma_x,
        width_y=sigma_y,
        frequency=sf,
        phase=2 * np.pi / 2,
        center=loc,
        grid_size=img_res,
        transformation_matrix=transformation_matrix,
    )

    if unit_norm:
        gb = gb / np.linalg.norm(gb)

    xx, yy = generate_gabor_filter_span_as_grid(
        center=np.array(loc),
        grid_size=img_res,
        transformation_matrix=transformation_matrix,
    )
    grid = xx, yy

    return ArbitraryNeuronModel(np.stack(gb[None])), grid


def odd_gabor_neuron_generator(
    loc, img_res, transformation_matrix=None, unit_norm=True
):
    sigma_x = 0.14
    sigma_y = 0.14
    sf = 2.3
    gb = generate_gabor_filter(
        theta=0,
        width_x=sigma_x,
        width_y=sigma_y,
        frequency=sf,
        phase=1 * np.pi / 2,
        center=loc,
        grid_size=img_res,
        transformation_matrix=transformation_matrix,
    )

    if unit_norm:
        gb = gb / np.linalg.norm(gb)

    xx, yy = generate_gabor_filter_span_as_grid(
        center=np.array(loc),
        grid_size=img_res,
        transformation_matrix=transformation_matrix,
    )
    grid = xx, yy

    return ArbitraryNeuronModel(np.stack(gb[None])), grid


def complex_neuron_generator(loc, img_res, transformation_matrix=None, unit_norm=True):
    sigma_x = 0.14
    sigma_y = 0.14
    sf = 2.3
    gb1 = generate_gabor_filter(
        theta=0,
        width_x=sigma_x,
        width_y=sigma_y,
        frequency=sf,
        phase=1 * np.pi / 2,
        center=loc,
        grid_size=img_res,
        transformation_matrix=transformation_matrix,
    )

    sigma_x = 0.14
    sigma_y = 0.14
    gb2 = generate_gabor_filter(
        theta=0,
        width_x=sigma_x,
        width_y=sigma_y,
        frequency=sf,
        phase=2 * np.pi / 2,
        center=loc,
        grid_size=img_res,
        transformation_matrix=transformation_matrix,
    )

    if unit_norm:
        gb1 = gb1 / np.linalg.norm(gb1)
        gb2 = gb2 / np.linalg.norm(gb2)

    xx, yy = generate_gabor_filter_span_as_grid(
        center=np.array(loc),
        grid_size=img_res,
        transformation_matrix=transformation_matrix,
    )
    grid = xx, yy

    return ComplexCell(gb1, gb2), grid



def simulated(model_type="demo1", img_res=[100, 100]):

    if model_type == "demo1":
        loc = [0.2, .2]
        neuron_model1, _ = phase_invariant_neuron_generator(loc, img_res, num_filters=20)
        loc = [-0.2, -.2]
        t_mat = np.array([[0.70710678, -0.70710678], [0.70710678,  0.70710678]])
        neuron_model2, _ = phase_invariant_neuron_generator(loc, img_res, transformation_matrix=t_mat, num_filters=20)
        loc = [-0.2, .2]
        # neuron_model3, _ = neuron3_generator(loc, img_res)
        neuron_model3, _ = neuron2_generator(loc, img_res, num_filters=60)
        return ArbitraryMultiNeuronModel([neuron_model1, neuron_model2, neuron_model3])
    
    else:
        raise ValueError(f"Model type {model_type} does not exist.")


import warnings

import torch
from torch import optim

from featurevis.exceptions import FeatureVisException


def gradient_ascent(f, x, transform=None, regularization=None, gradient_f=None,
                    post_update=None, optim_name='SGD', step_size=0.1, optim_kwargs={},
                    num_iterations=1000, save_iters=None, print_iters=100):
    """ Maximize f(x) via gradient ascent.

    Objective: f(transform(x)) - regularization(transform(x))
    Update: x_{t+1} = post_update(x_{t} + step_size * gradient_f(x_{t}.grad))

    Arguments:
        f (function): Real-valued differentiable function to be optimized
        x (torch.Tensor): Initial guess of the input to optimize.
        transform (function): Differentiable transformation applied to x before sending it
            through the model, e.g., an image generator, jittering, scaling, etc.
        regularization (function): Differentiable regularization term, e.g., natural
            prior, total variation, bilateral filters, etc.
        gradient_f (function): Non-differentiable. Receives the gradient of x and outputs
            a preconditioned gradient, e.g., blurring, masking, etc.
        post_update (function): Non-differentiable. Function applied to x after each
            gradient update, e.g., keep the image norm to some value, blurring, etc.
        optim_name (string): Optimizer to use: SGD or Adam.
        step_size (float): Size of the step size to give every iteration.
        optim_kwargs (dict): Dictionary with kwargs for the optimizer
        num_iterations (int): Number of gradient ascent steps.
        save_iters (None or int): How often to save x. If None, it returns the best x;
            otherwise it saves x after each save_iters iterations.
        print_iters (int): Print some results every print_iters iterations.

    Returns:
        optimal_x (torch.Tensor): x that maximizes the desired function. If save_iters is
            not None, this will be a list of tensors.
        fevals (list): Function evaluations at each iteration. We also evaluate at x_0
            (the original input) so this will have max_iterations + 1 elements.
        reg_terms (list): Value of the regularization term at each iteration. We also
            evaluate at x_0 (the original input) so this will have max_iterations + 1
            elements. Empty if regularization is None.

    Note:
        transform, regularization, gradient_f and post_update receive one positional
        parameter (its input) and the following optional named parameters:
            iteration (int): Current iteration (starts at 1).

        The number of optional parameters may increase so we recommend to write functions
        that receive **kwargs (or use the varargin decorator below) to make sure they will
        still work if we add other optional parameters in the future.
    """
    # Basic checks
    if x.dtype != torch.float32:
        raise ValueError('x must be of torch.float32 dtype')
    x = x.detach().clone()  # to avoid changing original
    x.requires_grad_()

    # Declare optimizer
    if optim_name == 'SGD':
        optimizer = optim.SGD([x], lr=step_size, **optim_kwargs)
    elif optim_name == 'Adam':
        optimizer = optim.Adam([x], lr=step_size, **optim_kwargs)
    else:
        raise ValueError("Expected optim_name to be 'SGD' or 'Adam'")

    # Run gradient ascent
    fevals = []  # to store function evaluations
    reg_terms = []  # to store regularization function evaluations
    saved_xs = []  # to store xs (ignored if save_iters is None)
    for i in range(1, num_iterations + 1):
        # Zero gradients
        if x.grad is not None:
            x.grad.zero_()

        # Transform input
        transformed_x = x if transform is None else transform(x, iteration=i)

        # f(x)
        feval = f(transformed_x)
        fevals.append(feval.item())

        # Regularization
        if regularization is not None:
            reg_term = regularization(transformed_x, iteration=i)
            reg_terms.append(reg_term.item())
        else:
            reg_term = 0

        # Compute gradient
        (-feval + reg_term).backward()
        if x.grad is None:
            raise FeatureVisException('Gradient did not reach x.')

        # Precondition gradient
        x.grad = x.grad if gradient_f is None else gradient_f(x.grad, iteration=i)
        if (torch.abs(x.grad) < 1e-9).all():
            warnings.warn('Gradient for x is all zero')

        # Gradient ascent step (on x)
        optimizer.step()

        # Cleanup
        if post_update is not None:
            with torch.no_grad():
                x[:] = post_update(x, iteration=i)  # in place so the optimizer still points to the right object

        # Report results
        if i % print_iters == 0:
            feval = feval.item()
            reg_term = reg_term if regularization is None else reg_term.item()
            x_std = x.std().item()
            print('Iter {}: f(x) = {:.2f}, reg(x) = {:.2f}, std(x) = {:.2f}'.format(i,
                feval, reg_term, x_std))

        # Save x
        if save_iters is not None and i % save_iters == 0:
            saved_xs.append(x.detach().clone())

    # Record f(x) and regularization(x) for the final x
    with torch.no_grad():
        transformed_x = x if transform is None else transform(x, iteration=i + 1)

        feval = f(transformed_x)
        fevals.append(feval.item())

        if regularization is not None:
            reg_term = regularization(transformed_x, iteration=i + 1)
            reg_terms.append(reg_term.item())
    print('Final f(x) = {:.2f}'.format(fevals[-1]))

    # Set opt_x
    opt_x = x.detach().clone() if save_iters is None else saved_xs

    return opt_x, fevals, reg_terms


import warnings
import numpy as np
import torch
import torch.nn as nn
from .utils.monkey_model import (
    se_core_full_gauss_readout,
    se_core_point_readout,
)
from neuralpredictors.layers.readouts import PointPooled2d, FullGaussian2d
from huggingface_hub import hf_hub_download

data_info = {
    "3631896544452": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 32,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3632669014376": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 21,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3632932714885": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 11,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3633364677437": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 10,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3634055946316": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 21,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3634142311627": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 14,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3634658447291": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 5,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3634744023164": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 12,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3635178040531": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 6,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3635949043110": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 10,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3636034866307": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 20,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3636552742293": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 24,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3637161140869": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 22,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3637248451650": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 7,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3637333931598": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 9,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3637760318484": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 20,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3637851724731": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 14,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3638367026975": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 16,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3638456653849": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 2,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3638885582960": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 5,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3638373332053": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 10,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3638541006102": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 17,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3638802601378": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 7,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3638973674012": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 18,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3639060843972": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 12,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3639406161189": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 14,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3640011636703": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 2,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3639664527524": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 18,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3639492658943": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 17,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3639749909659": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 8,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3640095265572": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 25,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
    "3631807112901": {
        "input_dimensions": torch.Size([128, 1, 93, 93]),
        "input_channels": 1,
        "output_dimension": 29,
        "img_mean": 114.54466,
        "img_std": 64.07781,
    },
}

pointpool_model_config = {
    "pad_input": False,
    "stack": -1,
    "depth_separable": True,
    "input_kern": 24,
    "gamma_input": 10,
    "gamma_readout": 0.5,
    "hidden_dilation": 2,
    "hidden_kern": 9,
    "se_reduction": 16,
    "n_se_blocks": 2,
    "hidden_channels": 32,
    "linear": False,
}

gaussian_model_config = {
    "pad_input": False,
    "stack": -1,
    "depth_separable": True,
    "input_kern": 24,
    "gamma_input": 10,
    "gamma_readout": 0.5,
    "hidden_dilation": 2,
    "hidden_kern": 9,
    "se_reduction": 16,
    "hidden_channels": 32,
    "linear": False,
    "n_se_blocks": 0,
}


def get_pointpool_monkey_model(
    seed=None, data_info=data_info, model_config=pointpool_model_config
):
    seed = seed if seed is not None else np.random.randint(0, 100)
    model = se_core_point_readout(seed=seed, data_info=data_info, **model_config)
    return model


class keyless_pointpool_monkey_model(nn.Module):
    def __init__(self, model_initial):
        super().__init__()
        model_initial.eval()
        keys = list(model_initial.readout.keys())
        att = dict(**model_initial.readout[keys[0]].__dict__)
        att["bias"] = True
        att["outdims"] = 458
        att["pool_steps"] = 2
        start = 0
        readout = PointPooled2d(**att)
        for k in keys:
            r = model_initial.readout[k]
            add = r.outdims
            end = start + add
            readout.features.data[:, :, :, start:end] = r.features.data
            readout.grid.data[:, start:end, :, :] = r.grid.data
            readout.bias.data[start:end] = r.bias.data
            start = start + add
        self.core = model_initial.core
        self.readout = readout
        self.eval()

    def forward(self, x):
        x = self.readout(self.core(x))
        return torch.nn.functional.elu(x) + 1


class ensamble_keyless_pointpool_monkey_model(nn.Module):
    def __init__(self, models_paths):
        super().__init__()
        models = []
        for path in models_paths:
            model = get_pointpool_monkey_model()
            model.load_state_dict(torch.load(path))
            models.append(model)
        self.models = nn.ModuleList(
            keyless_pointpool_monkey_model(model) for model in models
        )
        self.eval()

    def forward(self, x):
        x = torch.stack([model(x) for model in self.models]).mean(dim=0)
        return x


def get_gaussian_monkey_model(
    seed=None, data_info=data_info, model_config=gaussian_model_config
):
    seed = seed if seed is not None else np.random.randint(0, 100)
    model = se_core_full_gauss_readout(seed=seed, data_info=data_info, **model_config)
    return model


class keyless_gaussian_monkey_model(nn.Module):
    def __init__(self, model_initial):
        super().__init__()
        model_initial.eval()
        keys = list(model_initial.readout.keys())
        att = dict(**model_initial.readout[keys[0]].__dict__)
        att["bias"] = True
        att["outdims"] = 458
        start = 0
        readout = FullGaussian2d(**att)
        for k in keys:
            r = model_initial.readout[k]
            add = r.outdims
            end = start + add
            readout.features.data[:, :, :, start:end] = r.features.data
            readout.grid.data[:, start:end, :, :] = r.grid.data
            readout.bias.data[start:end] = r.bias.data
            start = start + add
        self.core = model_initial.core
        self.readout = readout
        self.eval()

    def forward(self, x):
        x = self.readout(self.core(x))
        return torch.nn.functional.elu(x) + 1


class ensamble_keyless_gaussian_monkey_model(nn.Module):
    def __init__(self, models_paths):
        super().__init__()
        models = []
        for path in models_paths:
            model = get_gaussian_monkey_model()
            model.load_state_dict(torch.load(path))
            models.append(model)
        self.models = nn.ModuleList(
            keyless_gaussian_monkey_model(model) for model in models
        )
        self.eval()

    def forward(self, x):
        x = torch.stack([model(x) for model in self.models]).mean(dim=0)
        return x


def monkey_v1(model_type="pointpool"):
    if model_type == "pointpool":
        # download model weights
        model_paths = [
            hf_hub_download(
                repo_id="mobashiri/bashiri_baroni_iclr2025",
                filename=f"monkey_pointpool_1.pt",
            ),
            hf_hub_download(
                repo_id="mobashiri/bashiri_baroni_iclr2025",
                filename=f"monkey_pointpool_2.pt",
            ),
            hf_hub_download(
                repo_id="mobashiri/bashiri_baroni_iclr2025",
                filename=f"monkey_pointpool_3.pt",
            ),
        ]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            return ensamble_keyless_pointpool_monkey_model(model_paths)
    elif model_type == "gaussian":
        # download model weights
        model_paths = [
            hf_hub_download(
                repo_id="mobashiri/bashiri_baroni_iclr2025",
                filename=f"monkey_gaussian_1.pt",
            ),
            hf_hub_download(
                repo_id="mobashiri/bashiri_baroni_iclr2025",
                filename=f"monkey_gaussian_2.pt",
            ),
            hf_hub_download(
                repo_id="mobashiri/bashiri_baroni_iclr2025",
                filename=f"monkey_gaussian_3.pt",
            ),
        ]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            return ensamble_keyless_gaussian_monkey_model(model_paths)
    else:
        raise ValueError("readout_type must be 'pointpool' or 'gaussian'")



import warnings

import numpy as np
import torch
from torch import nn
from torch.nn import Parameter
from torch.nn import functional as F

from .base import ConfigurationError, Readout


class Gaussian2d(Readout):
    """
    Instantiates an object that can used to learn a point in the core feature space for each neuron,
    sampled from a Gaussian distribution with some mean and variance at train but set to mean at test time, that best predicts its response.

    The readout receives the shape of the core as 'in_shape', the number of units/neurons being predicted as 'outdims', 'bias' specifying whether
    or not bias term is to be used and 'init_range' range for initialising the mean and variance of the gaussian distribution from which we sample to
    uniform distribution, U(-init_range,init_range) and  uniform distribution, U(0.0, 3*init_range) respectively.
    The grid parameter contains the normalized locations (x, y coordinates in the core feature space) and is clipped to [-1.1] as it a
    requirement of the torch.grid_sample function. The feature parameter learns the best linear mapping between the feature
    map from a given location, sample from Gaussian at train time but set to mean at eval time, and the unit's response with or without an additional elu non-linearity.

    Args:
        in_shape (list): shape of the input feature map [channels, width, height]
        outdims (int): number of output units
        bias (bool): adds a bias term
        init_mu_range (float): initialises the the mean with Uniform([-init_range, init_range])
                            [expected: positive value <=1]
        init_sigma_range (float): initialises sigma with Uniform([0.0, init_sigma_range]).
                It is recommended however to use a fixed initialization, for faster convergence.
                For this, set fixed_sigma to True.
        batch_sample (bool): if True, samples a position for each image in the batch separately
                            [default: True as it decreases convergence time and performs just as well]
        align_corners (bool): Keyword agrument to gridsample for bilinear interpolation.
                It changed behavior in PyTorch 1.3. The default of align_corners = True is setting the
                behavior to pre PyTorch 1.3 functionality for comparability.
        fixed_sigma (bool). Recommended behavior: True. But set to false for backwards compatibility.
                If true, initialized the sigma not in a range, but with the exact value given for all neurons.
    """

    def __init__(
        self,
        in_shape,
        outdims,
        bias,
        init_mu_range=0.5,
        init_sigma_range=0.5,
        batch_sample=True,
        align_corners=True,
        fixed_sigma=False,
        mean_activity=None,
        feature_reg_weight=1.0,
        gamma_readout=None,  # depricated, use feature_reg_weight instead
        **kwargs,
    ):
        warnings.warn(
            "Gaussian2d is deprecated and will be removed in the future. Use `layers.readout.NonIsoGaussian2d` instead",
            DeprecationWarning,
        )
        super().__init__()
        if init_mu_range > 1.0 or init_mu_range <= 0.0 or init_sigma_range <= 0.0:
            raise ValueError("either init_mu_range doesn't belong to [0.0, 1.0] or init_sigma_range is non-positive")
        self.in_shape = in_shape
        self.feature_reg_weight = self.resolve_deprecated_gamma_readout(feature_reg_weight, gamma_readout)
        c, w, h = in_shape
        self.outdims = outdims
        self.batch_sample = batch_sample
        self.grid_shape = (1, outdims, 1, 2)
        self.mu = Parameter(torch.Tensor(*self.grid_shape))  # mean location of gaussian for each neuron
        self.sigma = Parameter(torch.Tensor(*self.grid_shape))  # standard deviation for gaussian for each neuron
        self.features = Parameter(torch.Tensor(1, c, 1, outdims))  # feature weights for each channel of the core
        self.mean_activity = mean_activity
        if bias:
            bias = Parameter(torch.Tensor(outdims))
            self.register_parameter("bias", bias)
        else:
            self.register_parameter("bias", None)

        self.init_mu_range = init_mu_range
        self.init_sigma_range = init_sigma_range
        self.align_corners = align_corners
        self.fixed_sigma = fixed_sigma
        self.initialize(mean_activity)

    def initialize(self, mean_activity=None):
        """
        Initializes the mean, and sigma of the Gaussian readout along with the features weights
        """
        if mean_activity is None:
            mean_activity = self.mean_activity
        self.mu.data.uniform_(-self.init_mu_range, self.init_mu_range)
        if self.fixed_sigma:
            self.sigma.data.uniform_(self.init_sigma_range, self.init_sigma_range)
        else:
            self.sigma.data.uniform_(0, self.init_sigma_range)
            warnings.warn(
                "sigma is sampled from uniform distribuiton, instead of a fixed value. Consider setting "
                "fixed_sigma to True"
            )
        self.features.data.fill_(1 / self.in_shape[0])
        if self.bias is not None:
            self.initialize_bias(mean_activity=mean_activity)

    def sample_grid(self, batch_size, sample=None):
        """
        Returns the grid locations from the core by sampling from a Gaussian distribution
        Args:
            batch_size (int): size of the batch
            sample (bool/None): sample determines whether we draw a sample from Gaussian distribution, N(mu,sigma), defined per neuron
                            or use the mean, mu, of the Gaussian distribution without sampling.
                           if sample is None (default), samples from the N(mu,sigma) during training phase and
                             fixes to the mean, mu, during evaluation phase.
                           if sample is True/False, overrides the model_state (i.e training or eval) and does as instructed
        """
        with torch.no_grad():
            self.mu.clamp_(min=-1, max=1)  # at eval time, only self.mu is used so it must belong to [-1,1]
            self.sigma.clamp_(min=0)  # sigma/variance is always a positive quantity

        grid_shape = (batch_size,) + self.grid_shape[1:]

        sample = self.training if sample is None else sample

        if sample:
            norm = self.mu.new(*grid_shape).normal_()
        else:
            norm = self.mu.new(*grid_shape).zero_()  # for consistency and CUDA capability

        return torch.clamp(
            norm * self.sigma + self.mu, min=-1, max=1
        )  # grid locations in feature space sampled randomly around the mean self.mu

    @property
    def grid(self):
        return self.sample_grid(batch_size=1, sample=False)

    def feature_l1(self, reduction="sum", average=None):
        """
        Returns l1 regularization term for features.
        Args:
            average(bool): Deprecated (see reduction) if True, use mean of weights for regularization
            reduction(str): Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'
        """
        return self.apply_reduction(self.features.abs(), reduction=reduction, average=average)

    def regularizer(self, reduction="sum", average=None):
        return self.feature_l1(reduction=reduction, average=average) * self.feature_reg_weight

    def forward(self, x, sample=None, shift=None, out_idx=None, **kwargs):
        """
        Propagates the input forwards through the readout
        Args:
            x: input data
            sample (bool/None): sample determines whether we draw a sample from Gaussian distribution, N(mu,sigma), defined per neuron
                            or use the mean, mu, of the Gaussian distribution without sampling.
                           if sample is None (default), samples from the N(mu,sigma) during training phase and
                             fixes to the mean, mu, during evaluation phase.
                           if sample is True/False, overrides the model_state (i.e training or eval) and does as instructed
            shift (bool): shifts the location of the grid (from eye-tracking data)
            out_idx (bool): index of neurons to be predicted

        Returns:
            y: neuronal activity
        """
        N, c, w, h = x.size()
        c_in, w_in, h_in = self.in_shape
        if (c_in, w_in, h_in) != (c, w, h):
            raise ValueError("the specified feature map dimension is not the readout's expected input dimension")
        feat = self.features.view(1, c, self.outdims)
        bias = self.bias
        outdims = self.outdims

        if self.batch_sample:
            # sample the grid_locations separately per image per batch
            grid = self.sample_grid(batch_size=N, sample=sample)  # sample determines sampling from Gaussian
        else:
            # use one sampled grid_locations for all images in the batch
            grid = self.sample_grid(batch_size=1, sample=sample).expand(N, outdims, 1, 2)

        if out_idx is not None:
            if isinstance(out_idx, np.ndarray):
                if out_idx.dtype == bool:
                    out_idx = np.where(out_idx)[0]
            feat = feat[:, :, out_idx]
            grid = grid[:, out_idx]
            if bias is not None:
                bias = bias[out_idx]
            outdims = len(out_idx)

        if shift is not None:
            grid = grid + shift[:, None, None, :]

        y = F.grid_sample(x, grid, align_corners=self.align_corners)
        y = (y.squeeze(-1) * feat).sum(1).view(N, outdims)

        if self.bias is not None:
            y = y + bias
        return y

    def __repr__(self):
        c, w, h = self.in_shape
        r = self.__class__.__name__ + " (" + "{} x {} x {}".format(c, w, h) + " -> " + str(self.outdims) + ")"
        if self.bias is not None:
            r += " with bias"
        for ch in self.children():
            r += "  -> " + ch.__repr__() + "\n"
        return r


class FullGaussian2d(Readout):
    """
    A readout using a spatial transformer layer whose positions are sampled from one Gaussian per neuron. Mean
    and covariance of that Gaussian are learned.

    Args:
        in_shape (list, tuple): shape of the input feature map [channels, width, height]
        outdims (int): number of output units
        bias (bool): adds a bias term
        init_mu_range (float): initialises the the mean with Uniform([-init_range, init_range])
                            [expected: positive value <=1]. Default: 0.1
        init_sigma (float): The standard deviation of the Gaussian with `init_sigma` when `gauss_type` is
            'isotropic' or 'uncorrelated'. When `gauss_type='full'` initialize the square root of the
            covariance matrix with with Uniform([-init_sigma, init_sigma]). Default: 1
        batch_sample (bool): if True, samples a position for each image in the batch separately
                            [default: True as it decreases convergence time and performs just as well]
        align_corners (bool): Keyword agrument to gridsample for bilinear interpolation.
                It changed behavior in PyTorch 1.3. The default of align_corners = True is setting the
                behavior to pre PyTorch 1.3 functionality for comparability.
        gauss_type (str): Which Gaussian to use. Options are 'isotropic', 'uncorrelated', or 'full' (default).
        grid_mean_predictor (dict): Parameters for a predictor of the mean grid locations. Has to have a form like
                        {
                        'hidden_layers':0,
                        'hidden_features':20,
                        'final_tanh': False,
                        }
        shared_features (dict): Used when the feature vectors are shared (within readout between neurons) or between
                this readout and other readouts. Has to be a dictionary of the form
               {
                    'match_ids': (numpy.array),
                    'shared_features': torch.nn.Parameter or None
                }
                The match_ids are used to match things that should be shared within or across scans.
                If `shared_features` is None, this readout will create its own features. If it is set to
                a feature Parameter of another readout, it will replace the features of this readout. It will be
                access in increasing order of the sorted unique match_ids. For instance, if match_ids=[2,0,0,1],
                there should be 3 features in order [0,1,2]. When this readout creates features, it will do so in
                that order.
        shared_grid (dict): Like `shared_features`. Use dictionary like
               {
                    'match_ids': (numpy.array),
                    'shared_grid': torch.nn.Parameter or None
                }
                See documentation of `shared_features` for specification.

        source_grid (numpy.array):
                Source grid for the grid_mean_predictor.
                Needs to be of size neurons x grid_mean_predictor[input_dimensions]

    """

    def __init__(
        self,
        in_shape,
        outdims,
        bias,
        init_mu_range=0.1,
        init_sigma=1,
        batch_sample=True,
        align_corners=True,
        gauss_type="full",
        grid_mean_predictor=None,
        shared_features=None,
        shared_grid=None,
        source_grid=None,
        mean_activity=None,
        feature_reg_weight=1.0,
        gamma_readout=None,  # depricated, use feature_reg_weight instead
        **kwargs,
    ):

        super().__init__()
        self.feature_reg_weight = self.resolve_deprecated_gamma_readout(feature_reg_weight, gamma_readout)
        self.mean_activity = mean_activity
        # determines whether the Gaussian is isotropic or not
        self.gauss_type = gauss_type

        if init_mu_range > 1.0 or init_mu_range <= 0.0 or init_sigma <= 0.0:
            raise ValueError("either init_mu_range doesn't belong to [0.0, 1.0] or init_sigma_range is non-positive")

        # store statistics about the images and neurons
        self.in_shape = in_shape
        self.outdims = outdims

        # sample a different location per example
        self.batch_sample = batch_sample

        # position grid shape
        self.grid_shape = (1, outdims, 1, 2)

        # the grid can be predicted from another grid
        self._predicted_grid = False
        self._shared_grid = False
        self._original_grid = not self._predicted_grid

        if grid_mean_predictor is None and shared_grid is None:
            self._mu = Parameter(torch.Tensor(*self.grid_shape))  # mean location of gaussian for each neuron
        elif grid_mean_predictor is not None and shared_grid is not None:
            raise ConfigurationError("Shared grid_mean_predictor and shared_grid_mean cannot both be set")
        elif grid_mean_predictor is not None:
            self.init_grid_predictor(source_grid=source_grid, **grid_mean_predictor)
        elif shared_grid is not None:
            self.initialize_shared_grid(**(shared_grid or {}))

        if gauss_type == "full":
            self.sigma_shape = (1, outdims, 2, 2)
        elif gauss_type == "uncorrelated":
            self.sigma_shape = (1, outdims, 1, 2)
        elif gauss_type == "isotropic":
            self.sigma_shape = (1, outdims, 1, 1)
        else:
            raise ValueError(f'gauss_type "{gauss_type}" not known')

        self.init_sigma = init_sigma
        self.sigma = Parameter(torch.Tensor(*self.sigma_shape))  # standard deviation for gaussian for each neuron

        self.initialize_features(**(shared_features or {}))

        if bias:
            bias = Parameter(torch.Tensor(outdims))
            self.register_parameter("bias", bias)
        else:
            self.register_parameter("bias", None)

        self.init_mu_range = init_mu_range
        self.align_corners = align_corners
        self.initialize(mean_activity)

    @property
    def shared_features(self):
        return self._features

    @property
    def shared_grid(self):
        return self._mu

    @property
    def features(self):
        if self._shared_features:
            return self.scales * self._features[..., self.feature_sharing_index]
        else:
            return self._features

    @property
    def grid(self):
        return self.sample_grid(batch_size=1, sample=False)

    @property
    def mu_dispersion(self):
        """
        Returns the standard deviation of the learned positions.
        Is used as a regularizer to push neurons to learn similar positions.

        Returns:
            mu_dispersion(float): average dispersion of the mean 2d-position
        """

        return self._mu.squeeze().std(0).sum()

    def feature_l1(self, reduction="sum", average=None):
        """
        Returns l1 regularization term for features.
        Args:
            average(bool): Deprecated (see reduction) if True, use mean of weights for regularization
            reduction(str): Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'
        """
        if self._original_features:
            return self.apply_reduction(self.features.abs(), reduction=reduction, average=average)
        else:
            return 0

    def regularizer(self, reduction="sum", average=None):
        return self.feature_l1(reduction=reduction, average=average) * self.feature_reg_weight

    @property
    def mu(self):
        if self._predicted_grid:
            return self.mu_transform(self.source_grid.squeeze()).view(*self.grid_shape)
        elif self._shared_grid:
            if self._original_grid:
                return self._mu[:, self.grid_sharing_index, ...]
            else:
                return self.mu_transform(self._mu.squeeze())[self.grid_sharing_index].view(*self.grid_shape)
        else:
            return self._mu

    def sample_grid(self, batch_size, sample=None):
        """
        Returns the grid locations from the core by sampling from a Gaussian distribution
        Args:
            batch_size (int): size of the batch
            sample (bool/None): sample determines whether we draw a sample from Gaussian distribution, N(mu,sigma), defined per neuron
                            or use the mean, mu, of the Gaussian distribution without sampling.
                           if sample is None (default), samples from the N(mu,sigma) during training phase and
                             fixes to the mean, mu, during evaluation phase.
                           if sample is True/False, overrides the model_state (i.e training or eval) and does as instructed
        """
        with torch.no_grad():
            self.mu.clamp_(
                min=-1, max=1
            )  # at eval time, only self.mu is used so it must belong to [-1,1] # sigma/variance i    s always a positive quantity

        grid_shape = (batch_size,) + self.grid_shape[1:]

        sample = self.training if sample is None else sample
        if sample:
            norm = self.mu.new(*grid_shape).normal_()
        else:
            norm = self.mu.new(*grid_shape).zero_()  # for consistency and CUDA capability

        if self.gauss_type != "full":
            return torch.clamp(
                norm * self.sigma + self.mu, min=-1, max=1
            )  # grid locations in feature space sampled randomly around the mean self.mu
        else:
            return torch.clamp(
                torch.einsum("ancd,bnid->bnic", self.sigma, norm) + self.mu,
                min=-1,
                max=1,
            )  # grid locations in feature space sampled randomly around the mean self.mu

    def init_grid_predictor(self, source_grid, hidden_features=20, hidden_layers=0, final_tanh=False):
        self._original_grid = False
        layers = [nn.Linear(source_grid.shape[1], hidden_features if hidden_layers > 0 else 2)]

        for i in range(hidden_layers):
            layers.extend(
                [
                    nn.ELU(),
                    nn.Linear(hidden_features, hidden_features if i < hidden_layers - 1 else 2),
                ]
            )

        if final_tanh:
            layers.append(nn.Tanh())
        self.mu_transform = nn.Sequential(*layers)

        source_grid = source_grid - source_grid.mean(axis=0, keepdims=True)
        source_grid = source_grid / np.abs(source_grid).max()
        self.register_buffer("source_grid", torch.from_numpy(source_grid.astype(np.float32)))
        self._predicted_grid = True

    def initialize(self, mean_activity=None):
        """
        Initializes the mean, and sigma of the Gaussian readout along with the features weights
        """
        if mean_activity is None:
            mean_activity = self.mean_activity
        if not self._predicted_grid or self._original_grid:
            self._mu.data.uniform_(-self.init_mu_range, self.init_mu_range)

        if self.gauss_type != "full":
            self.sigma.data.fill_(self.init_sigma)
        else:
            self.sigma.data.uniform_(-self.init_sigma, self.init_sigma)
        self._features.data.fill_(1 / self.in_shape[0])
        if self._shared_features:
            self.scales.data.fill_(1.0)
        if self.bias is not None:
            self.initialize_bias(mean_activity=mean_activity)

    def initialize_features(self, match_ids=None, shared_features=None):
        """
        The internal attribute `_original_features` in this function denotes whether this instance of the FullGuassian2d
        learns the original features (True) or if it uses a copy of the features from another instance of FullGaussian2d
        via the `shared_features` (False). If it uses a copy, the feature_l1 regularizer for this copy will return 0
        """
        c, w, h = self.in_shape
        self._original_features = True
        if match_ids is not None:
            assert self.outdims == len(match_ids)

            n_match_ids = len(np.unique(match_ids))
            if shared_features is not None:
                assert shared_features.shape == (
                    1,
                    c,
                    1,
                    n_match_ids,
                ), f"shared features need to have shape (1, {c}, 1, {n_match_ids})"
                self._features = shared_features
                self._original_features = False
            else:
                self._features = Parameter(
                    torch.Tensor(1, c, 1, n_match_ids)
                )  # feature weights for each channel of the core
            self.scales = Parameter(torch.Tensor(1, 1, 1, self.outdims))  # feature weights for each channel of the core
            _, sharing_idx = np.unique(match_ids, return_inverse=True)
            self.register_buffer("feature_sharing_index", torch.from_numpy(sharing_idx))
            self._shared_features = True
        else:
            self._features = Parameter(
                torch.Tensor(1, c, 1, self.outdims)
            )  # feature weights for each channel of the core
            self._shared_features = False

    def initialize_shared_grid(self, match_ids=None, shared_grid=None):
        c, w, h = self.in_shape

        if match_ids is None:
            raise ConfigurationError("match_ids must be set for sharing grid")
        assert self.outdims == len(match_ids), "There must be one match ID per output dimension"

        n_match_ids = len(np.unique(match_ids))
        if shared_grid is not None:
            assert shared_grid.shape == (
                1,
                n_match_ids,
                1,
                2,
            ), f"shared grid needs to have shape (1, {n_match_ids}, 1, 2)"
            self._mu = shared_grid
            self._original_grid = False
            self.mu_transform = nn.Linear(2, 2)
            self.mu_transform.bias.data.fill_(0.0)
            self.mu_transform.weight.data = torch.eye(2)
        else:
            self._mu = Parameter(torch.Tensor(1, n_match_ids, 1, 2))  # feature weights for each channel of the core
        _, sharing_idx = np.unique(match_ids, return_inverse=True)
        self.register_buffer("grid_sharing_index", torch.from_numpy(sharing_idx))
        self._shared_grid = True

    def forward(self, x, sample=None, shift=None, out_idx=None, **kwargs):
        """
        Propagates the input forwards through the readout
        Args:
            x: input data
            sample (bool/None): sample determines whether we draw a sample from Gaussian distribution, N(mu,sigma), defined per neuron
                            or use the mean, mu, of the Gaussian distribution without sampling.
                           if sample is None (default), samples from the N(mu,sigma) during training phase and
                             fixes to the mean, mu, during evaluation phase.
                           if sample is True/False, overrides the model_state (i.e training or eval) and does as instructed
            shift (bool): shifts the location of the grid (from eye-tracking data)
            out_idx (bool): index of neurons to be predicted

        Returns:
            y: neuronal activity
        """
        N, c, w, h = x.size()
        c_in, w_in, h_in = self.in_shape
        if (c_in, w_in, h_in) != (c, w, h):
            warnings.warn("the specified feature map dimension is not the readout's expected input dimension")
        feat = self.features.view(1, c, self.outdims)
        bias = self.bias
        outdims = self.outdims

        if self.batch_sample:
            # sample the grid_locations separately per image per batch
            grid = self.sample_grid(batch_size=N, sample=sample)  # sample determines sampling from Gaussian
        else:
            # use one sampled grid_locations for all images in the batch
            grid = self.sample_grid(batch_size=1, sample=sample).expand(N, outdims, 1, 2)

        if out_idx is not None:
            if isinstance(out_idx, np.ndarray):
                if out_idx.dtype == bool:
                    out_idx = np.where(out_idx)[0]
            feat = feat[:, :, out_idx]
            grid = grid[:, out_idx]
            if bias is not None:
                bias = bias[out_idx]
            outdims = len(out_idx)

        if shift is not None:
            grid = grid + shift[:, None, None, :]

        y = F.grid_sample(x, grid, align_corners=self.align_corners)
        y = (y.squeeze(-1) * feat).sum(1).view(N, outdims)

        if self.bias is not None:
            y = y + bias
        return y

    def __repr__(self):
        c, w, h = self.in_shape
        r = self.gauss_type + " "
        r += self.__class__.__name__ + " (" + "{} x {} x {}".format(c, w, h) + " -> " + str(self.outdims) + ")"
        if self.bias is not None:
            r += " with bias"
        if self._shared_features:
            r += ", with {} features".format("original" if self._original_features else "shared")

        if self._predicted_grid:
            r += ", with predicted grid"
        if self._shared_grid:
            r += ", with {} grid".format("original" if self._original_grid else "shared")

        for ch in self.children():
            r += "  -> " + ch.__repr__() + "\n"
        return r


class RemappedGaussian2d(FullGaussian2d):
    """
    A readout using a spatial transformer layer whose positions are sampled from one Gaussian per neuron. Mean
    and covariance of that Gaussian are learned. In addition, there is an image dependent remapping of neurons
    locations.

    For most parameters see:  FullGaussian2d

    Args:
        remap_layers (int): number of layers of the remapping network
        remap_kernel (int): conv kernel size of the remapping network
        max_remap_amplitude (int): maximal amplitude of remapping (factor on output of remapping network)


    """

    def __init__(self, *args, remap_layers=2, remap_kernel=3, max_remap_amplitude=0.2, **kwargs):

        super().__init__(*args, **kwargs)
        channels, width, height = self.in_shape
        remapper = nn.Sequential()
        for i in range(remap_layers - 1):
            remapper.add_module(f"conv{i}", nn.Conv2d(channels, channels, remap_kernel, padding=True))
            remapper.add_module(f"norm{i}", nn.BatchNorm2d(channels))
            remapper.add_module(f"nonlin{i}", nn.ELU())
        else:
            remapper.add_module(
                f"conv{remap_layers}",
                nn.Conv2d(channels, 2, remap_kernel, padding=True),
            )
            remapper.add_module(f"norm{remap_layers}", nn.BatchNorm2d(2))
            remapper.add_module(f"nonlin{remap_layers}", nn.Tanh())
        self.remap_field = remapper
        self.max_remap_amplitude = max_remap_amplitude

    @staticmethod
    def init_conv(m):
        if isinstance(m, nn.Conv2d):
            nn.init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0)

    def initialize_remap_field(self):
        self.apply(self.init_conv)

    def initialize(self, mean_activity=None, **kwargs):
        super().initialize(mean_activity)
        self.initialize_remap_field()

    def forward(self, x, sample=None, shift=None, out_idx=None, **kwargs):
        offset_field = self.remap_field(x) * self.max_remap_amplitude

        N, c, w, h = x.size()
        c_in, w_in, h_in = self.in_shape
        if (c_in, w_in, h_in) != (c, w, h):
            raise ValueError("the specified feature map dimension is not the readout's expected input dimension")
        feat = self.features.view(1, c, self.outdims)
        bias = self.bias
        outdims = self.outdims

        if self.batch_sample:
            # sample the grid_locations separately per image per batch
            grid = self.sample_grid(batch_size=N, sample=sample)  # sample determines sampling from Gaussian
        else:
            # use one sampled grid_locations for all images in the batch
            grid = self.sample_grid(batch_size=1, sample=sample).expand(N, outdims, 1, 2)

        if out_idx is not None:
            if isinstance(out_idx, np.ndarray):
                if out_idx.dtype == bool:
                    out_idx = np.where(out_idx)[0]
            feat = feat[:, :, out_idx]
            grid = grid[:, out_idx]
            if bias is not None:
                bias = bias[out_idx]
            outdims = len(out_idx)

        offsets = F.grid_sample(offset_field, grid, align_corners=self.align_corners)
        grid = grid + offsets.permute(0, 2, 3, 1)
        if shift is not None:
            grid = grid + shift[:, None, None, :]

        y = F.grid_sample(x, grid, align_corners=self.align_corners)
        y = (y.squeeze(-1) * feat).sum(1).view(N, outdims)


class DeterministicGaussian2d(Readout):
    """
    'DeterministicGaussian2d' class instantiates an object that can be used to learn a
    Normal distribution in the core feature space for each neuron, this is
    applied to the pixel grid to give a simple weighting.

    The variance should be decreased over training to achieve localization.
    The readout receives the shape of the core as 'in_shape', the number of units/neurons being predicted as
    'outdims', 'bias' specifying whether or not bias term is to be used.

    From the mean and and variance, a filter unique for each outdim is built, that is of the 2d density of
    the normal distribution. The filter is applied only on the spatial dimensions of the input.

    Args:
        in_shape (list): shape of the input feature map [channels, width, height]
        outdims (int): number of output units
        bias (bool): adds a bias term
        positive (bool): if True, all weights will be constrained to have positive values, default: False
        constrain_mode (str): ['default', 'abs', 'elu'] sets the way the weights are constrained, default: 'default'

    Written by Claudius Gruner.
    """

    def __init__(
        self,
        in_shape,
        outdims,
        bias,
        init_mu_range=0.1,
        init_sigma=1,
        positive=False,
        constrain_mode="default",
        mean_activity=None,
        feature_reg_weight=1.0,
        gamma_readout=None,  # depricated, use feature_reg_weight instead
    ):

        super().__init__()

        if init_mu_range > 1.0 or init_mu_range <= 0.0 or init_sigma <= 0.0:
            raise ValueError("either init_mu_range doesn't belong to [0.0, 1.0] or init_sigma_range is non-positive")

        self.in_shape = in_shape
        c, w, h = in_shape
        self.outdims = outdims
        self.positive = positive
        self.feature_reg_weight = self.resolve_deprecated_gamma_readout(feature_reg_weight, gamma_readout)
        self.mean_activity = mean_activity
        if constrain_mode not in ["default", "abs", "elu"]:
            raise ValueError(
                "Value of parameter constrain_mode = "
                + str(constrain_mode)
                + " is invalid. Value must be in ['default', 'abs', 'elu'] "
            )
        self.constrain_mode = constrain_mode

        self.mu = Parameter(data=torch.zeros(outdims, 2), requires_grad=True)
        self.log_var = Parameter(
            data=torch.zeros(
                outdims,
            ),
            requires_grad=True,
        )
        self.grid = torch.nn.Parameter(data=self.make_mask_grid(), requires_grad=False)

        self.features = Parameter(torch.Tensor(c, outdims))  # saliency  weights for each channel from core

        if bias:
            bias = Parameter(torch.Tensor(outdims))
            self.register_parameter("bias", bias)
        else:
            self.register_parameter("bias", None)

        self.init_mu_range = init_mu_range
        self.init_sigma = init_sigma

        self.initialize(mean_activity)

    def make_mask_grid(self):
        xx, yy = torch.meshgrid(
            [
                torch.linspace(-1, 1, self.in_shape[1]),
                torch.linspace(-1, 1, self.in_shape[2]),
            ], indexing='ij'
        )
        grid = torch.stack([xx, yy], 2)[None, ...]
        return grid.repeat([self.outdims, 1, 1, 1])

    def mask(self, shift=None):

        self.mu.clamp_(min=-1, max=1)

        variances = torch.exp(self.log_var).view(-1, 1, 1)

        if shift is None:
            mu = self.mu
        else:
            mu = self.mu + shift[None, ...]
        mean = mu.view(self.outdims, 1, 1, -1)
        pdf = self.grid - mean
        pdf = torch.sum(pdf**2, dim=-1) / variances
        pdf = torch.exp(-0.5 * pdf)
        # normalize to sum=1
        pdf = pdf / torch.sum(pdf, dim=(1, 2), keepdim=True)
        return pdf

    def initialize(self, mean_activity=None):
        """
        initialize function initializes the mean, sigma for the Gaussian readout and features weights
        """
        if mean_activity is None:
            mean_activity = self.mean_activity
        self.mu.data.uniform_(-self.init_mu_range, self.init_mu_range)
        self.log_var.data.fill_(np.log(self.init_sigma**2))
        self.features.data.fill_(1 / self.in_shape[0])
        if self.bias is not None:
            self.initialize_bias(mean_activity=mean_activity)

    def feature_l1(self, reduction="sum", average=None):
        """
        Returns l1 regularization term for features.
        Args:
            average(bool): Deprecated (see reduction) if True, use mean of weights for regularization
            reduction(str): Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'
        """
        return self.apply_reduction(self.features.abs(), reduction=reduction, average=average)

    def variance_l1(self, reduction="sum", average=None):
        """
        variance_l1 function returns the l1 regularization term either the mean or just the sum of weights
        Args:
            average(bool): if True, use mean of weights for regularization
            reduction(str): Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'
        """
        return self.apply_reduction(torch.exp(self.log_var), reduction=reduction, average=average)

    def regularizer(self, reduction="sum", average=None):
        return (
            self.feature_l1(reduction=reduction, average=average)
            + self.variance_l1(reduction=reduction, average=average)
        ) * self.feature_reg_weight

    def forward(self, x, shift=None, out_idx=None, **kwargs):
        N, c, w, h = x.size()
        feat = self.features

        if out_idx is None:
            grid = self.grid
            bias = self.bias
            outdims = self.outdims
        else:
            feat = feat[:, out_idx]
            grid = self.grid[:, out_idx]
            if self.bias is not None:
                bias = self.bias[out_idx]
            outdims = len(out_idx)

        if shift is None:
            mask = self.mask()
        else:
            mask = self.mask(shift=shift)

        if self.positive:
            if self.constrain_mode == "default":
                mask.data.clamp_min_(0)
                feat.data.clamp_min_(0)
            elif self.constrain_mode == "abs":
                mask = torch.abs(mask)
                feat = torch.abs(feat)
            elif self.constrain_mode == "elu":
                mask = F.elu(mask) + 1
                feat = F.elu(feat) + 1

        y = torch.einsum("bcij,nij,cn->bn", x, mask, feat)

        if self.bias is not None:
            y = y + bias
        return y

    def __repr__(self):
        c, w, h = self.in_shape
        r = self.__class__.__name__ + " (" + "{} x {} x {}".format(c, w, h) + " -> " + str(self.outdims) + ")"
        if self.bias is not None:
            r += " with bias"
        for ch in self.children():
            r += "  -> " + ch.__repr__() + "\n"
        return r


class Gaussian3d(Readout):
    """
    This readout instantiates an object that can used to learn a point in the core feature space for each neuron,
    sampled from a Gaussian distribution with some mean and variance at train but set to mean at test time, that best predicts its response.

    The readout receives the shape of the core as 'in_shape', the number of units/neurons being predicted as 'outdims', 'bias' specifying whether
    or not bias term is to be used and 'init_range' range for initialising the mean and variance of the gaussian distribution from which we sample to
    uniform distribution, U(-init_mu_range,init_mu_range) and  uniform distribution, U(0.0, init_sigma_range) respectively.
    The grid parameter contains the normalized locations (x, y coordinates in the core feature space) and is clipped to [-1.1] as it a
    requirement of the torch.grid_sample function. The feature parameter learns the best linear mapping between the feature
    map from a given location, sample from Gaussian at train time but set to mean at eval time, and the unit's response with or without an additional elu non-linearity.

    Args:
        in_shape (list): shape of the input feature map [channels, width, height]
        outdims (int): number of output units
        bias (bool): adds a bias term
        init_mu_range (float): initialises the the mean with Uniform([-init_range, init_range])
                            [expected: positive value <=1]
        init_sigma_range (float): initialises sigma with Uniform([0.0, init_sigma_range]).
                It is recommended however to use a fixed initialization, for faster convergence.
                For this, set fixed_sigma to True.
        batch_sample (bool): if True, samples a position for each image in the batch separately
                            [default: True as it decreases convergence time and performs just as well]
        align_corners (bool): Keyword agrument to gridsample for bilinear interpolation.
                It changed behavior in PyTorch 1.3. The default of align_corners = True is setting the
                behavior to pre PyTorch 1.3 functionality for comparability.
        fixed_sigma (bool). Recommended behavior: True. But set to false for backwards compatibility.
                If true, initialized the sigma not in a range, but with the exact value given for all neurons.
    """

    def __init__(
        self,
        in_shape,
        outdims,
        bias,
        init_mu_range=0.5,
        init_sigma_range=0.5,
        batch_sample=True,
        align_corners=True,
        fixed_sigma=False,
        mean_activity=None,
        feature_reg_weight=1.0,
        gamma_readout=None,  # depricated, use feature_reg_weight instead
        **kwargs,
    ):
        super().__init__()
        if init_mu_range > 1.0 or init_mu_range <= 0.0 or init_sigma_range <= 0.0:
            raise ValueError("init_mu_range or init_sigma_range is not within required limit!")
        self.in_shape = in_shape
        self.outdims = outdims
        self.feature_reg_weight = self.resolve_deprecated_gamma_readout(feature_reg_weight, gamma_readout)
        self.batch_sample = batch_sample
        self.grid_shape = (1, 1, outdims, 1, 3)
        self.mu = Parameter(torch.Tensor(*self.grid_shape))  # mean location of gaussian for each neuron
        self.sigma = Parameter(torch.Tensor(*self.grid_shape))  # standard deviation for gaussian for each neuron
        self.features = Parameter(torch.Tensor(1, 1, 1, outdims))  # saliency weights for each channel from core
        self.mean_activity = mean_activity
        if bias:
            bias = Parameter(torch.Tensor(outdims))
            self.register_parameter("bias", bias)
        else:
            self.register_parameter("bias", None)

        self.init_mu_range = init_mu_range
        self.init_sigma_range = init_sigma_range
        self.align_corners = align_corners
        self.fixed_sigma = fixed_sigma
        self.initialize(mean_activity)

    def sample_grid(self, batch_size, sample=None):
        """
        Returns the grid locations from the core by sampling from a Gaussian distribution
        Args:
            batch_size (int): size of the batch
            sample (bool/None): sample determines whether we draw a sample from Gaussian distribution, N(mu,sigma), defined per neuron
                            or use the mean, mu, of the Gaussian distribution without sampling.
                           if sample is None (default), samples from the N(mu,sigma) during training phase and
                             fixes to the mean, mu, during evaluation phase.
                           if sample is True/False, overrides the model_state (i.e training or eval) and does as instructed
        """

        with torch.no_grad():
            self.mu.clamp_(min=-1, max=1)  # at eval time, only self.mu is used so it must belong to [-1,1]
            self.sigma.clamp_(min=0)  # sigma/variance is always a positive quantity

        grid_shape = (batch_size,) + self.grid_shape[1:]

        sample = self.training if sample is None else sample

        if sample:
            norm = self.mu.new(*grid_shape).normal_()
        else:
            norm = self.mu.new(*grid_shape).zero_()  # for consistency and CUDA capability

        return torch.clamp(
            norm * self.sigma + self.mu, min=-1, max=1
        )  # grid locations in feature space sampled randomly around the mean self.mu

    @property
    def grid(self):
        return self.sample_grid(batch_size=1, sample=False)

    def initialize(self, mean_activity=None):
        if mean_activity is None:
            mean_activity = self.mean_activity
        self.mu.data.uniform_(-self.init_mu_range, self.init_mu_range)
        if self.fixed_sigma:
            self.sigma.data.uniform_(self.init_sigma_range, self.init_sigma_range)
        else:
            self.sigma.data.uniform_(0, self.init_sigma_range)
            warnings.warn(
                "sigma is sampled from uniform distribuiton, instead of a fixed value. Consider setting "
                "fixed_sigma to True"
            )
        self.features.data.fill_(1 / self.in_shape[0])
        if self.bias is not None:
            self.initialize_bias(mean_activity=mean_activity)

    def regularizer(self, reduction="sum", average=None):
        raise NotImplementedError(
            "The regularizer for this readout needs to be implemented! See issue https://github.com/sinzlab/neuralpredictors/issues/127"
        )

    def forward(self, x, sample=None, shift=None, out_idx=None, **kwargs):
        """
        Propagates the input forwards through the readout
        Args:
            x: input data
            sample (bool/None): sample determines whether we draw a sample from Gaussian distribution, N(mu,sigma), defined per neuron
                            or use the mean, mu, of the Gaussian distribution without sampling.
                           if sample is None (default), samples from the N(mu,sigma) during training phase and
                             fixes to the mean, mu, during evaluation phase.
                           if sample is True/False, overrides the model_state (i.e training or eval) and does as instructed
            shift (bool): shifts the location of the grid (from eye-tracking data)
            out_idx (bool): index of neurons to be predicted

        Returns:
            y: neuronal activity
        """
        N, c, w, h = x.size()
        c_in, w_in, h_in = self.in_shape
        if (c_in, w_in, h_in) != (c, w, h):
            raise ValueError("the specified feature map dimension is not the readout's expected input dimension")
        x = x.view(N, 1, c, w, h)
        feat = self.features
        bias = self.bias
        outdims = self.outdims

        if self.batch_sample:
            # sample the grid_locations separately per image per batch
            grid = self.sample_grid(batch_size=N, sample=sample)  # sample determines sampling from Gaussian
        else:
            # use one sampled grid_locations for all images in the batch
            grid = self.sample_grid(batch_size=1, sample=sample).expand(N, outdims, 1, 3)

        if out_idx is not None:
            # out_idx specifies the indices to subset of neurons for training/testing
            if isinstance(out_idx, np.ndarray):
                if out_idx.dtype == bool:
                    out_idx = np.where(out_idx)[0]
            feat = feat[:, :, :, out_idx]
            grid = grid[:, :, out_idx]
            if bias is not None:
                bias = bias[out_idx]
            outdims = len(out_idx)

        if shift is not None:
            grid = grid + shift[:, None, None, :]

        y = F.grid_sample(x, grid, align_corners=self.align_corners)
        y = (y.squeeze(-1) * feat).sum(1).view(N, outdims)

        if self.bias is not None:
            y = y + bias
        return y


class UltraSparse(Readout):
    """
    This readout instantiates an object that can used to learn one or more features (with or without
    a shared mean in the x-y plane) in the core feature space for each neuron, sampled from a Gaussian distribution
    with some mean and variance at training but set to mean at test time, that best predicts its response.

    The readout receives the shape of the core as 'in_shape', the number of units/neurons being predicted as 'outdims', 'bias' specifying whether
    or not bias term is to be used and 'init_range' range for initialising the mean and variance of the gaussian distribution from which we sample to
    uniform distribution, U(-init_mu_range,init_mu_range) and  uniform distribution, U(0.0, init_sigma_range) respectively.
    The grid parameter contains the normalized locations (x, y coordinates in the core feature space) and is clipped to [-1.1] as it a
    requirement of the torch.grid_sample function. The feature parameter learns the best linear mapping between the feature
    map from a given location, sample from Gaussian at train time but set to mean at eval time, and the unit's response with or without an additional elu non-linearity.

    Args:
        in_shape (list): shape of the input feature map [channels, width, height]
        outdims (int): number of output units
        bias (bool): adds a bias term
        init_mu_range (float): initialises the the mean with Uniform([-init_range, init_range])
                            [expected: positive value <=1]
        init_sigma_range (float): initialises sigma with Uniform([0.0, init_sigma_range])
        batch_sample (bool): if True, samples a position for each image in the batch separately
                            [default: True as it decreases convergence time and performs just as well]
        num_filters (int): number of points in the core-features to be learned for each neuron
                           [default: 1, an instance of sparsest readout]
        shared_mean (bool): if True, the mean in the x-y plane (image-plane) is shared across all channels
                           [default: False]

        align_corners (bool): Keyword agrument to gridsample for bilinear interpolation.
                It changed behavior in PyTorch 1.3. The default of align_corners = True is setting the
                behavior to pre PyTorch 1.3 functionality for comparability.
        fixed_sigma (bool). Recommended behavior: True. But set to false for backwards compatibility.
                If true, initialized the sigma not in a range, but with the exact value given for all neurons.
    """

    def __init__(
        self,
        in_shape,
        outdims,
        bias,
        init_mu_range,
        init_sigma_range,
        batch_sample=True,
        num_filters=1,
        shared_mean=False,
        align_corners=True,
        fixed_sigma=False,
        mean_activity=None,
        feature_reg_weight=1.0,
        gamma_readout=None,  # depricated, use feature_reg_weight instead
        **kwargs,
    ):

        super().__init__()
        if init_mu_range > 1.0 or init_mu_range <= 0.0 or init_sigma_range <= 0.0:
            raise ValueError("either init_mu_range doesn't belong to [0.0, 1.0] or init_sigma_range is non-positive!")
        self.in_shape = in_shape
        c, w, h = in_shape
        self.outdims = outdims
        self.feature_reg_weight = self.resolve_deprecated_gamma_readout(feature_reg_weight, gamma_readout)
        self.batch_sample = batch_sample
        self.num_filters = num_filters
        self.shared_mean = shared_mean
        self.grid_shape = (1, 1, outdims * num_filters, 1, 3)
        self.mean_activity = mean_activity
        if shared_mean:

            self.gridxy_shape = (1, 1, outdims, 1, 2)
            self.gridch_shape = (1, 1, outdims * num_filters, 1, 1)
            self.mu_xy = Parameter(
                torch.Tensor(*self.gridxy_shape)
            )  # mean location (in xy dim) of gaussian for each neuron
            self.mu_ch = Parameter(
                torch.Tensor(*self.gridch_shape)
            )  # mean location (in ch dim) of gaussian for each neuron
            self.sigma_xy = Parameter(
                torch.Tensor(*self.gridxy_shape)
            )  # standard deviation for gaussian for each neuron
            self.sigma_ch = Parameter(torch.Tensor(*self.gridch_shape))

        else:

            self.mu = Parameter(torch.Tensor(*self.grid_shape))  # mean location of gaussian for each neuron
            self.sigma = Parameter(torch.Tensor(*self.grid_shape))  # standard deviation for gaussian for each neuron

        self.features = Parameter(
            torch.Tensor(1, 1, outdims, num_filters)
        )  # saliency  weights for each channel from core

        if bias:
            bias = Parameter(torch.Tensor(outdims))
            self.register_parameter("bias", bias)
        else:
            self.register_parameter("bias", None)

        self.init_mu_range = init_mu_range
        self.init_sigma_range = init_sigma_range
        self.align_corners = align_corners
        self.fixed_sigma = fixed_sigma
        self.initialize(mean_activity)

    def sample_grid(self, batch_size, sample=None):
        """
        Returns the grid locations from the core by sampling from a Gaussian distribution
        Args:
            batch_size (int): size of the batch
            sample (bool/None): sample determines whether we draw a sample from Gaussian distribution, N(mu,sigma), defined per neuron
                            or use the mean, mu, of the Gaussian distribution without sampling.
                           if sample is None (default), samples from the N(mu,sigma) during training phase and
                             fixes to the mean, mu, during evaluation phase.
                           if sample is True/False, overrides the model_state (i.e training or eval) and does as instructed
        """

        if self.shared_mean:
            # sample an xy location and keep it same across all filter channels
            # explicit clamping of mu and sigma along the channel dimension was needed as the clamping post cat was not working
            with torch.no_grad():
                self.mu_ch.clamp_(min=-1, max=1)  # at eval time, only self.mu is used so it must belong to [-1,1]
                self.sigma_ch.clamp_(min=0)  # sigma/variance is always a positive quantity
            self.mu = torch.cat((self.mu_xy.repeat(1, 1, self.num_filters, 1, 1), self.mu_ch), 4)
            self.sigma = torch.cat((self.sigma_xy.repeat(1, 1, self.num_filters, 1, 1), self.sigma_ch), 4)

        with torch.no_grad():
            self.mu.clamp_(min=-1, max=1)
            self.sigma.clamp_(min=0)

        grid_shape = (batch_size,) + self.grid_shape[1:]

        sample = self.training if sample is None else sample

        if sample:
            norm = self.mu.new(*grid_shape).normal_()
        else:
            norm = self.mu.new(*grid_shape).zero_()  # for consistency and CUDA capability

        return torch.clamp(
            norm * self.sigma + self.mu, min=-1, max=1
        )  # grid locations in feature space sampled randomly around the mean self.mu

    @property
    def grid(self):
        return self.sample_grid(batch_size=1, sample=False)

    def feature_l1(self, reduction="sum", average=None):
        """
        Returns l1 regularization term for features.
        Args:
            average(bool): Deprecated (see reduction) if True, use mean of weights for regularization
            reduction(str): Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'
        """
        return self.apply_reduction(self.features.abs(), reduction=reduction, average=average)

    def regularizer(self, reduction="sum", average=None):
        return (
            self.feature_l1(reduction=reduction, average=average)
            + self.variance_l1(reduction=reduction, average=average)
        ) * self.feature_reg_weight

    def initialize(self, mean_activity=None):
        if mean_activity is None:
            mean_activity = self.mean_activity
        if self.shared_mean:
            # initialise mu and sigma separately for xy and channel dimension.
            self.mu_ch.data.uniform_(-1, 1)
            self.mu_xy.data.uniform_(-self.init_mu_range, self.init_mu_range)

            if self.fixed_sigma:
                self.sigma_ch.data.uniform_(self.init_sigma_range, self.init_sigma_range)
                self.sigma_xy.data.uniform_(self.init_sigma_range, self.init_sigma_range)
            else:
                self.sigma_ch.data.uniform_(0, self.init_sigma_range)
                self.sigma_xy.data.uniform_(0, self.init_sigma_range)
                warnings.warn(
                    "sigma is sampled from uniform distribuiton, instead of a fixed value. Consider setting "
                    "fixed_sigma to True"
                )

        else:
            # initialise mu and sigma for x,y and channel dimensions.
            self.mu.data.uniform_(-self.init_mu_range, self.init_mu_range)
            self.sigma.data.uniform_(0, self.init_sigma_range)

        self.features.data.fill_(1 / self.in_shape[0])
        if self.bias is not None:
            self.initialize_bias(mean_activity=mean_activity)

    def forward(self, x, sample=True, shift=None, out_idx=None, **kwargs):
        """
        Propagates the input forwards through the readout
        Args:
            x: input data
            sample (bool/None): sample determines whether we draw a sample from Gaussian distribution, N(mu,sigma), defined per neuron
                            or use the mean, mu, of the Gaussian distribution without sampling.
                           if sample is None (default), samples from the N(mu,sigma) during training phase and
                             fixes to the mean, mu, during evaluation phase.
                           if sample is True/False, overrides the model_state (i.e training or eval) and does as instructed
            shift (bool): shifts the location of the grid (from eye-tracking data)
            out_idx (bool): index of neurons to be predicted

        Returns:
            y: neuronal activity
        """

        N, c, w, h = x.size()
        c_in, w_in, h_in = self.in_shape
        if (c_in, w_in, h_in) != (c, w, h):
            raise ValueError("the specified feature map dimension is not the readout's expected input dimension")
        x = x.view(N, 1, c, w, h)
        feat = self.features
        bias = self.bias
        outdims = self.outdims

        if self.batch_sample:
            # sample the grid_locations separately per image per batch
            grid = self.sample_grid(batch_size=N, sample=sample)  # sample determines sampling from Gaussian
        else:
            # use one sampled grid_locations for all images in the batch
            grid = self.sample_grid(batch_size=1, sample=sample).expand(N, 1, outdims * self.num_filters, 1, 3)

        if out_idx is not None:
            # predict output only for neurons given by out_idx
            if isinstance(out_idx, np.ndarray):
                if out_idx.dtype == bool:
                    out_idx = np.where(out_idx)[0]
            feat = feat[:, :, :, out_idx]
            grid = grid[:, :, out_idx]
            if bias is not None:
                bias = bias[out_idx]
            outdims = len(out_idx)

        if shift is not None:  # it might not be valid now but have kept it for future devop.
            grid = grid + shift[:, None, None, :]

        y = F.grid_sample(x, grid, align_corners=self.align_corners).squeeze(-1)
        z = y.view((N, 1, self.num_filters, outdims)).permute(0, 1, 3, 2)  # reorder the dims
        z = torch.einsum(
            "nkpf,mkpf->np", z, feat
        )  # dim: batch_size, 1, num_neurons, num_filters -> batch_size, num_neurons

        if self.bias is not None:
            z = z + bias
        return z

    def __repr__(self):
        c, w, h = self.in_shape
        r = self.__class__.__name__ + " (" + "{} x {} x {}".format(c, w, h) + " -> " + str(self.outdims) + ")"
        if self.bias is not None:
            r += " with bias"
        for ch in self.children():
            r += "  -> " + ch.__repr__() + "\n"
        return r



import logging

import numpy as np
import torch
from torch import nn as nn
from torch.nn import functional as F

logger = logging.getLogger(__name__)


def elu1(x):
    return F.elu(x, inplace=True) + 1.0


class Elu1(nn.Module):
    """
    Elu activation function shifted by 1 to ensure that the
    output stays positive. That is:
    Elu1(x) = Elu(x) + 1
    """

    def forward(self, x):
        return elu1(x)


def log1exp(x):
    return torch.log(1.0 + torch.exp(x))


class Log1Exp(nn.Module):
    def forward(self, x):
        return log1exp(x)


def adaptive_elu(x, xshift, yshift):
    return F.elu(x - xshift, inplace=True) + yshift


class AdaptiveELU(nn.Module):
    """
    ELU shifted by user specified values. This helps to ensure the output to stay positive.
    """

    def __init__(self, xshift, yshift, **kwargs):
        super(AdaptiveELU, self).__init__(**kwargs)

        self.xshift = xshift
        self.yshift = yshift

    def forward(self, x):
        return adaptive_elu(x, self.xshift, self.yshift)


class PiecewiseLinearExpNonlinearity(nn.Module):
    def __init__(
        self,
        number_of_neurons,
        bias=False,
        initial_value=0.01,
        vmin=-3,
        vmax=6,
        num_bins=50,
        smooth_reg_weight=0,
        smoothnes_reg_order=2,
    ):
        super().__init__()

        self.bias = bias
        self.initial = initial_value
        self.vmin = vmin
        self.vmax = vmax
        self.neurons = number_of_neurons
        self.smooth_reg_weight = smooth_reg_weight
        self.smoothnes_reg_order = smoothnes_reg_order

        self.num_bins = 2 * int(num_bins / 2)

        if self.bias:
            self.b = torch.nn.Parameter(torch.empty((number_of_neurons,), dtype=torch.float32).fill_(self.initial))
        self.a = torch.nn.Parameter(torch.empty((self.num_bins, self.neurons), dtype=torch.float32).fill_(0))

        bins = np.linspace(self.vmin, self.vmax, self.num_bins + 1, endpoint=True).reshape(1, -1)
        bins_mtx = np.tile(bins, [1, self.neurons, 1])
        bins_mtx = np.transpose(bins_mtx, (0, 2, 1)).astype(np.float32)
        # shape: 1, num_bins, num_neurons
        self.bins = torch.nn.Parameter(torch.from_numpy(bins_mtx), requires_grad=False)

        self.zero = torch.nn.Parameter(torch.zeros((1,), dtype=torch.float32), requires_grad=False)

    def tent(self, x, a, b):
        return torch.min(torch.max(self.zero, x - a), torch.max(self.zero, 2 * b - a - x)) / (b - a)

    def linstep(self, x, a, b):
        return torch.min(b - a, torch.max(self.zero, x - a)) / (b - a)

    def smoothness_regularizer(self, verbose=False):
        penalty = 0
        kernel = torch.tensor(np.reshape([-1.0, 1.0], (1, 1, 2)), dtype=torch.float32).cuda()

        w = torch.reshape(self.a, (-1, 1, self.num_bins))  # shape: neurons, 1, bins
        for k in range(self.smoothnes_reg_order):
            w = F.conv1d(w, kernel)
            penalty += torch.sum(torch.mean(w**2, 1))
        penalty = torch.sum(self.smooth_reg_weight * penalty)
        if verbose:
            logger.info(f"PieceWiseLinearExpNonLin, Smoothness penalty: {penalty}")
        return penalty

    def forward(self, x):
        if self.bias:
            # a bias is added
            x = x + self.b

        g = torch.nn.functional.elu(x - 1) + 1
        xx = torch.reshape(x, (-1, 1, self.neurons))

        # a tent function is applied on the data in multiple bins.
        tents = self.tent(xx, self.bins[:, :-2, :], self.bins[:, 1:-1, :])
        linstep = self.linstep(xx, self.bins[:, -2:-1, :], self.bins[:, -1:, :])

        t = torch.cat((tents, linstep), dim=1)
        # bins shape: 1, num_bins, num_neurons
        # t shape: batch, bins, neurons

        h = torch.sum(torch.exp(self.a) * t, dim=1)

        return g * h

    def visualize(self, vmin=None, vmax=None, iters=1000, show=True, return_fig=False, neurons=range(10)):
        if vmin is None:
            vmin = self.vmin - 1
        if vmax is None:
            vmax = self.vmax + 1

        inpts = torch.from_numpy(np.tile(np.linspace(vmin, vmax, iters).astype(np.float32), [self.neurons, 1]).T).cuda()
        outs = self.forward(inpts)

        f = plt.figure()
        ax = f.add_subplot(1, 1, 1)
        ax.plot(inpts.cpu().detach().numpy()[:, neurons], outs.cpu().detach().numpy()[:, neurons])
        ax.set_xlabel("Response before alteration")
        ax.set_ylabel("Response after alteration")

        plt.grid(which="both")

        if show:
            f.show()
        if return_fig:
            return f


class MultiplePiecewiseLinearExpNonlinearity(nn.ModuleDict):
    def __init__(
        self,
        n_neurons_dict,
        bias=False,
        initial_value=0.01,
        vmin=-3,
        vmax=6,
        num_bins=50,
        smooth_reg_weight=0,
        smoothnes_reg_order=2,
    ):
        # super init to get the _module attribute
        super().__init__()
        for k in n_neurons_dict:
            n_neurons = n_neurons_dict[k]
            self.add_module(
                k,
                PiecewiseLinearExpNonlinearity(
                    number_of_neurons=n_neurons,
                    bias=bias,
                    initial_value=initial_value,
                    vmin=vmin,
                    vmax=vmax,
                    num_bins=num_bins,
                    smooth_reg_weight=smooth_reg_weight,
                    smoothnes_reg_order=smoothnes_reg_order,
                ),
            )

    def forward(self, *args, data_key=None, **kwargs):
        if data_key is None and len(self) == 1:
            data_key = list(self.keys())[0]
        return self[data_key](*args, **kwargs)

    def regularizer(self, data_key):
        return self[data_key].smoothness_regularizer()



import logging
import warnings

import torch
from torch import nn

logger = logging.getLogger(__name__)


class Corr(nn.Module):
    def __init__(self, eps=1e-12, detach_target=True):
        """
        Compute correlation between the output and the target

        Args:
            eps (float, optional): Used to offset the computed variance to provide numerical stability.
                Defaults to 1e-12.
            detach_target (bool, optional): If True, `target` tensor is detached prior to computation. Appropriate when
                using this as a loss to train on. Defaults to True.
        """
        self.eps = eps
        self.detach_target = detach_target
        super().__init__()

    def forward(self, output, target):
        if self.detach_target:
            target = target.detach()
        delta_out = output - output.mean(0, keepdim=True)
        delta_target = target - target.mean(0, keepdim=True)

        var_out = delta_out.pow(2).mean(0, keepdim=True)
        var_target = delta_target.pow(2).mean(0, keepdim=True)

        corrs = (delta_out * delta_target).mean(0, keepdim=True) / (
            (var_out + self.eps) * (var_target + self.eps)
        ).sqrt()
        return corrs


class AvgCorr(nn.Module):
    def __init__(self, eps=1e-12):
        self.eps = eps
        super().__init__()

    def forward(self, output, target):
        # Add target detach here

        delta_out = output - output.mean(0, keepdim=True)
        delta_target = target - target.mean(0, keepdim=True)

        var_out = delta_out.pow(2).mean(0, keepdim=True)
        var_target = delta_target.pow(2).mean(0, keepdim=True)

        corrs = (delta_out * delta_target).mean(0, keepdim=True) / (
            (var_out + self.eps) * (var_target + self.eps)
        ).sqrt()
        # TODO: address the sign flip applied here
        return -corrs.mean()


class PoissonLoss(nn.Module):
    def __init__(self, bias=1e-12, per_neuron=False, avg=True):
        """
        Computes Poisson loss between the output and target. Loss is evaluated by computing log likelihood
        (up to a constant offset dependent on the target) that
        output prescribes the mean of the Poisson distribution and target is a sample from the distribution.

        Args:
            bias (float, optional): Value used to numerically stabilize evalution of the log-likelihood. This value is effecitvely added to the output during evaluation. Defaults to 1e-12.
            per_neuron (bool, optional): If set to True, the average/total Poisson loss is returned for each entry of the last dimension (assumed to be enumeration neurons) separately. Defaults to False.
            avg (bool, optional): If set to True, return mean loss. Otherwise returns the sum of loss. Defaults to True.
        """
        super().__init__()
        self.bias = bias
        self.per_neuron = per_neuron
        self.avg = avg
        if self.avg:
            warnings.warn("Poissonloss is averaged per batch. It's recommended to use `sum` instead")

    def forward(self, output, target):
        target = target.detach()
        loss = output - target * torch.log(output + self.bias)
        if not self.per_neuron:
            return loss.mean() if self.avg else loss.sum()
        else:
            loss = loss.view(-1, loss.shape[-1])
            return loss.mean(dim=0) if self.avg else loss.sum(dim=0)


class PoissonLoss3d(PoissonLoss):
    """
    Same as PoissonLoss, except that this automatically adjusts the length of the
    target along the 1st dimension (expected to correspond to the temporal dimension), such that
    when lag = target.size(1) - outout.size(1) > 0,
    PoissonLoss(output, target[:, lag:])
    is evaluted instead (thus equivalent to skipping the first `lag` frames).

    The constructor takes in the same arguments as in PoissonLoss
    """

    def forward(self, output, target):
        lag = target.size(1) - output.size(1)
        return super().forward(output, target[:, lag:])


class ExponentialLoss(nn.Module):
    def __init__(self, bias=1e-12, target_bias=1e-6, per_neuron=False):
        super().__init__()
        self.bias = bias
        self.target_bias = target_bias
        self.per_neuron = per_neuron

    def forward(self, output, target):
        output = output + self.bias

        target = (target + self.target_bias).detach()

        loss = target / output + torch.log(output)

        if not self.per_neuron:
            return loss.mean()
        else:
            return loss.view(-1, loss.shape[-1]).mean(dim=0)


class AnscombeMSE(nn.Module):
    def __init__(self, per_neuron=False):
        super().__init__()
        self.per_neuron = per_neuron

    @staticmethod
    def A(x):
        return 2 * torch.sqrt(x + 3 / 8)

    def forward(self, output, target):
        target = self.A(target).detach()
        output = self.A(output) - 1 / (4 * output.sqrt())

        loss = (target - output).pow(2)

        if not self.per_neuron:
            return loss.mean()
        else:
            return loss.view(-1, loss.shape[-1]).mean(dim=0)



import collections
from typing import TypeVar

T = TypeVar("T", bound=collections.abc.Mapping)


def deep_update(d: T, u: collections.abc.Mapping) -> T:
    """
    Recursively update a dictionary with the values from another dictionary
    Args:
        d: dictionary to be updated
        u: dictionary that is used to update `d`

    Returns:
        updated dictionary `d`
    """
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d



"""
This package contains all the components that are used to construct system identification models.
This includes:
  - cores: Modules used for extracting nonlinearly computed features shared among neurons.
  - readouts: Linear layers added on top of the core output to predict neurons' responses.
  - encoders: Classes which fuse core, readout, shifter (optional), and modulator (optional) into a single model.
  - shifters: Modules used for shifting (i.e. spatially) the features computed via core.
  - modulators: Modules used to incorporate the effect of other variables (e.g. behavioral variables) in response prediction.
  - special layers (e.g. DepthSeparable Convolution) and activation functions.
"""

from .affine import Bias2DLayer, Scale2DLayer
from .conv import DepthSeparableConv2d



from torch import nn


class FiringRateEncoder(nn.Module):
    def __init__(self, core, readout, *, shifter=None, modulator=None, elu_offset=0.0):
        """
        An Encoder that wraps the core, readout and optionally a shifter amd modulator into one model.
        The output is one positive value that can be interpreted as a firing rate, for example for a Poisson distribution.
        Args:
            core (nn.Module): Core model. Refer to neuralpredictors.layers.cores
            readout (nn.ModuleDict): MultiReadout model. Refer to neuralpredictors.layers.readouts
            elu_offset (float): Offset value in the final elu non-linearity. Defaults to 0.
            shifter (optional[nn.ModuleDict]): Shifter network. Refer to neuralpredictors.layers.shifters. Defaults to None.
            modulator (optional[nn.ModuleDict]): Modulator network. Modulator networks are not implemented atm (24/06/2021). Defaults to None.
        """
        super().__init__()
        self.core = core
        self.readout = readout
        self.shifter = shifter
        self.modulator = modulator
        self.offset = elu_offset

    def forward(
        self,
        inputs,
        *args,
        targets=None,
        data_key=None,
        behavior=None,
        pupil_center=None,
        trial_idx=None,
        shift=None,
        detach_core=False,
        **kwargs
    ):
        x = self.core(inputs)
        if detach_core:
            x = x.detach()

        if self.shifter:
            if pupil_center is None:
                raise ValueError("pupil_center is not given")
            shift = self.shifter[data_key](pupil_center, trial_idx)

        x = self.readout(x, data_key=data_key, shift=shift, **kwargs)

        if self.modulator:
            if behavior is None:
                raise ValueError("behavior is not given")
            x = self.modulator[data_key](x, behavior=behavior)

        return nn.functional.elu(x + self.offset) + 1

    def regularizer(self, data_key=None, reduction="sum", average=None, detach_core=False):
        reg = self.core.regularizer().detach() if detach_core else self.core.regularizer()
        reg += self.readout.regularizer(data_key=data_key, reduction=reduction, average=average)
        if self.shifter:
            reg += self.shifter.regularizer(data_key=data_key)
        if self.modulator:
            reg += self.modulator.regularizer(data_key=data_key)
        return reg
